"""
This is a script for the backend of the CPPPlot.feature() method.
"""
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib as mpl

import aaanalysis.utils as ut

from .utils_feature import get_list_parts, get_df_parts_, get_feature_matrix_, get_amino_acids_
from ._utils_cpp_plot_elements import PlotElements

COL_FEAT_VAL = "feature_values"


# I Helper Functions
# Data processing
def _get_df_feat_vals(feature=None, df_seq=None, df_scales=None, accept_gaps=False,
                      jmd_n_len=10, jmd_c_len=10):
    """Get DataFrame with values of selected feature for proteins in df"""
    list_parts = get_list_parts(features=feature)
    df_parts = get_df_parts_(df_seq=df_seq, list_parts=list_parts,
                             jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len)
    X = get_feature_matrix_(features=feature,
                            df_parts=df_parts,
                            df_scales=df_scales,
                            accept_gaps=accept_gaps,
                            n_jobs=1)
    if ut.COL_NAME in df_seq:
        list_names = df_seq[ut.COL_NAME].to_list()
    else:
        list_names = [f"Protein {i}" for i in range(len(df_seq))]
    df_feat_vals = pd.DataFrame({COL_FEAT_VAL: X.flatten(), ut.COL_NAME: list_names})
    return df_feat_vals


def _get_mean_vals_(df_feat_vals=None, labels=None, label_test=1, label_ref=0):
    """Get mean values for test and reference group and their difference"""
    mask_test = [x == label_test for x in labels]
    mask_ref = [x == label_ref for x in labels]
    mean_test = df_feat_vals[COL_FEAT_VAL][mask_test].mean()
    mean_ref = df_feat_vals[COL_FEAT_VAL][mask_ref].mean()
    mean_dif = round(mean_test - mean_ref, 3)
    return mean_test, mean_ref, mean_dif


def _get_df_show(df_feat_vals=None, feature=None, df_seq=None, names_to_show=None, show_seq=False,
                 jmd_n_len=10, jmd_c_len=10):
    """Get DataFrame with values to show"""
    names_to_show = [] if names_to_show is None else names_to_show
    df_show = df_feat_vals[df_feat_vals[ut.COL_NAME].isin(names_to_show)][[COL_FEAT_VAL, ut.COL_NAME, ut.COL_LABEL]]
    names_to_show = df_show[ut.COL_NAME].to_list()
    if show_seq and len(names_to_show) > 0:
        df_seq = df_seq[df_seq[ut.COL_NAME].isin(names_to_show)]
        list_parts = ["jmd_n", "tmd", "jmd_c"]
        df_parts = get_df_parts_(df_seq=df_seq, list_parts=list_parts,
                                 jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len)
        names_with_seq = []
        for i, (_, row) in enumerate(df_parts.iterrows()):
            jmd_n, tmd, jmd_c = row[list_parts]
            list_aa = get_amino_acids_(features=feature, jmd_n_seq=jmd_n, tmd_seq=tmd, jmd_c_seq=jmd_c)
            name = str(names_to_show[i]) + "\n" + list_aa[0]
            names_with_seq.append(name)
        df_show[ut.COL_NAME] = names_with_seq
    return df_show


# Additional plot information
def _add_mean_dif(ax, mean_test, mean_ref, mean_dif, y_mean_dif, alpha_dif):
    """Add area to highlight mean differences between reference and test group"""
    color_dif = ut.get_color_dif(mean_dif=mean_dif)
    ax.plot([mean_test, mean_ref], [y_mean_dif, y_mean_dif], "-", color=color_dif, linewidth=4)
    args_lines = dict(color="black", markeredgewidth=5)
    ax.plot(mean_ref, y_mean_dif, "|", **args_lines)
    ax.plot(mean_test, y_mean_dif, "|", **args_lines)

    ax.plot([mean_test, mean_test], [0, y_mean_dif], "-", color=color_dif, linewidth=2, alpha=0.25)
    ax.plot([mean_ref, mean_ref], [0, y_mean_dif], "-", color=color_dif, linewidth=2, alpha=0.25)

    bar = mpl.patches.Rectangle((mean_ref, 0), width=mean_dif, height=y_mean_dif, linewidth=1, color=color_dif,
                                zorder=0, clip_on=False, alpha=alpha_dif)

    ax.add_patch(bar)
    ax.plot([mean_test, mean_test], [0, y_mean_dif], "--", color="black", linewidth=2, alpha=1)
    ax.plot([mean_ref, mean_ref], [0, y_mean_dif], "--", color="black", linewidth=2, alpha=1)


def _add_mean_dif_text(ax, mean_test, mean_ref, mean_dif, y_mean_dif, name_test, name_ref,
                       color_test, color_ref, fontsize_name_test, fontsize_name_ref, fontsize_mean_dif):
    """Add text annotations for mean difference"""
    # Get default font sizes
    fs = ut.plot_gco()
    fontsize_mean_dif = fs + 2 if fontsize_mean_dif is None else fontsize_mean_dif
    fontsize_name_test = fs if fontsize_name_test is None else fontsize_name_test
    fontsize_name_ref = fs if fontsize_name_ref is None else fontsize_name_ref
    # Plot text
    middle = min([mean_test, mean_ref]) + abs(mean_dif) / 2
    str_mean_dif = f"{ut.LABEL_MEAN_DIF}={mean_dif:.3f}"  # Ensure mean_dif is formatted nicely
    ax.text(middle, y_mean_dif * 1.05, str_mean_dif, color="black", ha="center", size=fontsize_mean_dif)
    d = 0.015
    x_test = mean_test - d if mean_dif < 0 else mean_test + d
    ha_test = "right" if mean_dif < 0 else "left"
    x_ref = mean_ref - d if mean_dif > 0 else mean_ref + d
    ha_ref = "right" if mean_dif > 0 else "left"
    ax.text(x_test, y_mean_dif, name_test, ha=ha_test, va="center_baseline",
            color=color_test, size=fontsize_name_test)
    ax.text(x_ref, y_mean_dif, name_ref, ha=ha_ref, va="center_baseline",
            color=color_ref, size=fontsize_name_ref)


def _add_names_to_show(ax, df_show, name_test, color_test, color_ref, fontsize_names_to_show, show_seq):
    """Add individual data points with arrows to plot"""
    # Get default font sizes
    fs = ut.plot_gco()
    fontsize_names_to_show = fs - 2 if fontsize_names_to_show is None else fontsize_names_to_show
    for i, row in df_show.iterrows():
        val, name, label = row.values
        color = color_test if label == name_test else color_ref
        fontname = ut.FONT_AA if show_seq else None
        ax.annotate(name, xy=(val, 0), xytext=(0, 30), textcoords='offset points',
                    ha="center", va="center", fontname=fontname,
                    xycoords="data", fontsize=fontsize_names_to_show,
                    arrowprops=dict(arrowstyle="->", lw=3, color=color, shrinkA=0, shrinkB=0),
                    color="black")


# II Main Functions
def plot_feature(ax=None, figsize=(5.6, 4.8), feature=str,
                 df_seq=None, labels=None, label_test=1, label_ref=0,
                 df_scales=None, accept_gaps=None,
                 jmd_n_len=10, jmd_c_len=10,
                 name_test="TEST", name_ref="REF", names_to_show=None,
                 show_seq=False,
                 color_test="tab:green", color_ref="tab:gray",
                 histplot=False, alpha_hist=0.1, alpha_dif=0.25,
                 fontsize_mean_dif=15, fontsize_name_test=13,
                 fontsize_name_ref=13, fontsize_names_to_show=13,
                 ):
    """Plot distributions of feature values for test and reference datasets highlighting their mean difference."""
    # Get data
    df_feat_vals = _get_df_feat_vals(feature=feature, df_seq=df_seq, df_scales=df_scales, accept_gaps=accept_gaps,
                                     jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len)
    df_feat_vals[ut.COL_LABEL] = [name_test if x == label_test else name_ref for x in labels]
    mean_test, mean_ref, mean_dif = _get_mean_vals_(df_feat_vals=df_feat_vals, labels=labels,
                                                    label_test=label_test, label_ref=label_ref)
    # Get data for individual proteins to show
    df_show = _get_df_show(df_feat_vals=df_feat_vals, feature=feature, df_seq=df_seq,
                           names_to_show=names_to_show, show_seq=show_seq,
                           jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len)
    # Plotting
    pe = PlotElements()
    fig, ax = pe.set_figsize(ax=ax, figsize=figsize)
    args = dict(data=df_feat_vals, x=COL_FEAT_VAL,  hue=ut.COL_LABEL, ax=ax,
                palette=[color_test, color_ref], hue_order=[name_test, name_ref],
                legend=False)
    if histplot:
        sns.histplot(**args)
    else:
        sns.kdeplot(**args, fill=True, alpha=alpha_hist, common_norm=False)
    if ax is None:
        ax = plt.gca()
    fs = ut.plot_gco(option='font.size')  # Either None or current font size (if not provided, default size is used)
    # Set limits and labels using ax
    y_mean_dif = ax.get_ylim()[1] * 1.05
    y_max = max(y_mean_dif * 1.2, 1)
    ax.set_ylim(0, y_max)
    ax.set_xlim(0, 1)
    # Add mean dif
    _add_mean_dif(ax, mean_test, mean_ref, mean_dif, y_mean_dif, alpha_dif)
    _add_mean_dif_text(ax, mean_test, mean_ref, mean_dif, y_mean_dif, name_test, name_ref,
                       color_test, color_ref, fontsize_name_test, fontsize_name_ref, fontsize_mean_dif)
    # Add plot labels
    y_label = ut.LABEL_HIST_COUNT if histplot else ut.LABEL_HIST_DEN
    ax.set_ylabel(y_label, size=fs)
    ax.set_xlabel(ut.LABEL_FEAT_VAL, size=fs)
    # Add individual data points
    if len(df_show) != 0:
        _add_names_to_show(ax, df_show, name_test, color_test, color_ref, fontsize_names_to_show, show_seq)
    # Update ticks
    pe.set_x_ticks(ax=ax, fs=fs)
    pe.set_y_ticks(ax=ax, fs=fs)
    # Adjust plot
    sns.despine()
    plt.tight_layout()
    return ax




"""
This is the common utility script for the backend of the CPPPlot.heatmap() and
CPPPlot.feat_map() methods.
"""
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

import aaanalysis.utils as ut

from ._utils_cpp_plot_elements import PlotElements
from ._utils_cpp_plot_positions import PlotPartPositions


WIDTH_NORM_FACTOR = 110


# I Helper Functions
def _get_value_type(col_val="abs_auc"):
    """Determine the value type based on column value."""
    dict_val_type = ut.DICT_VALUE_TYPE
    if ut.COL_FEAT_IMPACT in col_val:
        val_type = "sum"
    elif ut.COL_MEAN_DIF in col_val:
        val_type = "mean"
    else:
        val_type = dict_val_type.get(col_val, "mean")
    return val_type


def _adjust_df_feat(df_feat=None, col_val=None, cbar_pct=True):
    """Adjusts feature values in `df_feat` based on percentage scaling and sets the limits for difference columns."""
    df_feat = df_feat.copy()
    val_type = _get_value_type(col_val=col_val)
    if val_type == "sum":
        sum_val = df_feat[col_val].abs().sum()
        adjust_col_val = sum_val < 1
    else:
        min_val = df_feat[col_val].min()
        max_val = df_feat[col_val].max()
        n_th = 1 if max_val < 0 or 0 < min_val else 2
        adjust_col_val = max(df_feat[col_val]) - min(df_feat[col_val]) < n_th
    if cbar_pct and adjust_col_val:
        df_feat[col_val] *= 100
    return df_feat


def _get_vmin_vmax(df_pos=None, vmin=None, vmax=None):
    """Obtain minimum and maximum values for the heatmap."""
    if vmax is not None and vmin is not None:
        return vmin, vmax   # Skip adjustment
    vmin = df_pos.min().min() if vmin is None else vmin
    vmax = df_pos.max().max() if vmax is None else vmax
    # Create centered vmin and vmax
    val = max((abs(vmin), abs(vmax)))
    vmax = val if vmax > 0 else 0
    vmin = -val if vmin < 0 else 0
    return vmin, vmax


def _get_bar_width(fig=None, len_seq=None):
    """Get consistent bar width to for category bars"""
    width, height = fig.get_size_inches()
    width_factor = 1 / width * 8
    bar_width = len_seq / WIDTH_NORM_FACTOR * width_factor
    return bar_width


# Get colormap
def _get_cmap_heatmap(df_pos=None, cmap=None, n_colors=101, facecolor_dark=True, vmin=None, vmax=None):
    """Create a sequential or diverging colormap for a heatmap based on data properties."""
    n_colors = n_colors if n_colors is not None else 101
    v_max = df_pos.max().max() if vmax is None else vmax
    v_min = df_pos.min().min() if vmin is None else vmin
    only_neg = v_max <= 0
    only_pos = v_min >= 0
    args = dict(n_colors=n_colors, facecolor_dark=facecolor_dark,
                only_neg=only_neg, only_pos=only_pos)
    cmap = ut.STR_CMAP_CPP if cmap is None else cmap
    cmap = ut.plot_get_cmap_(cmap=cmap, **args)
    return cmap


# Get and set colorbar arguments
def _get_cbar_ticks_heatmap(df_pos=None):
    """Calculate colorbar ticks for the heatmap."""
    stop_legend = df_pos.values.max()
    if type(stop_legend) is int:
        # Ticks for count datasets
        cbar_ticks = [int(x) for x in np.linspace(1, stop_legend, num=stop_legend)]
    else:
        cbar_ticks = None
    return cbar_ticks


def _get_cbar_args_heatmap(df_pos=None, cbar_kws=None):
    """Create colorbar arguments for the heatmap."""
    cbar_ticksize = ut.plot_gco() - 2
    cbar_ticks = _get_cbar_ticks_heatmap(df_pos=df_pos)
    dict_cbar = {"ticksize": cbar_ticksize, "labelsize": cbar_ticksize, "labelweight": "normal"}
    cbar_kws_ = dict(orientation="horizontal", ticks=cbar_ticks)
    if cbar_kws is not None:
        # Catch colorbar arguments that must be set manually
        for arg in dict_cbar:
            if arg in cbar_kws:
                dict_cbar[arg] = cbar_kws.pop(arg)
        cbar_kws_.update(**cbar_kws)
    return dict_cbar, cbar_kws_


def _set_cbar_heatmap(ax=None, dict_cbar=None, cbar_kws=None,
                      vmin=None, vmax=None, cbar_pct=True, weight="normal", fontsize=11):
    """Configure the heatmap's colorbar."""
    cbar = ax.collections[0].colorbar
    cbar.ax.tick_params(labelsize=dict_cbar["ticksize"])
    if "label" in cbar_kws:
        cbar.set_label(label=cbar_kws["label"], weight="bold", size=dict_cbar["labelsize"])
    cbar.ax.yaxis.label.set_size(dict_cbar["labelsize"])
    cbar_ticks = [x for x in cbar.get_ticks() if vmin <= x <= vmax]
    cbar.set_ticks(cbar_ticks)
    str_zero = "[0]"
    str_pct = "%" if cbar_pct else ""
    f = lambda x: int(x) if int(x) == float(x) else round(x, 1)
    cbar.set_ticklabels([f"{f(x)}{str_pct}" if float(x) != 0 else str_zero for x in cbar_ticks])
    if "label" in cbar_kws:
        cbar.set_label(label=cbar_kws["label"], weight=weight, size=fontsize)
    cbar.ax.xaxis.set_ticks_position('top')
    cbar.ax.xaxis.set_label_position('top')

    # Customization for the thin line on top or right of the colorbar
    orientation = cbar_kws.get("orientation", None)
    for spine in cbar.ax.spines.values():
        spine.set_visible(False)  # Show only the top spine
    if orientation is not None:
        pos = "top" if orientation == "horizontal" else "right"
        cbar.ax.spines[pos].set_visible(True)
        xtick_width = ut.plot_gco(option="xtick.major.width")
        cbar.ax.spines[pos].set_linewidth(xtick_width)


# II Main Functions
# Inner plotting function
def _plot_inner_heatmap(df_pos=None, ax=None, figsize=(8, 8),
                        start=1, tmd_len=20, jmd_n_len=10, jmd_c_len=10,
                        grid_linewidth=0.01, grid_linecolor=None,
                        border_linewidth=2,
                        facecolor_dark=True, vmin=None, vmax=None,
                        cbar_ax=None, cmap=None, cbar_kws=None,
                        x_shift=0.0,
                        xtick_size=11.0, xtick_width=2.0, xtick_length=None):
    """Plot the inner heatmap with specific settings and styles."""
    # Heatmap arguments
    vmin, vmax = _get_vmin_vmax(df_pos=df_pos, vmin=vmin, vmax=vmax)
    center = 0
    facecolor = "black" if facecolor_dark else "white"
    grid_linecolor = grid_linecolor if grid_linecolor is not None else  "gray" if facecolor_dark else "black"
    # Plot with 0 set to NaN
    pe = PlotElements()
    fig, ax = pe.set_figsize(ax=ax, figsize=figsize)
    data = df_pos.replace(0, np.NaN)
    ax = sns.heatmap(data, ax=ax,
                     center=center, vmin=vmin, vmax=vmax,
                     linewidths=grid_linewidth, linecolor=grid_linecolor,
                     cbar_ax=cbar_ax, cbar_kws=cbar_kws, cmap=cmap,
                     yticklabels=True, xticklabels=True)
    # Set default x-ticks
    pp = PlotPartPositions(tmd_len=tmd_len, jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len, start=start)
    pp.add_xticks(ax=ax, xticks_position="bottom", x_shift=x_shift,
                  xtick_size=xtick_size, xtick_width=xtick_width, xtick_length=xtick_length)
    # Remove y-ticks
    ax.tick_params(axis='y', which='both', length=0, labelsize=0)
    # Set frame
    for _, spine in ax.spines.items():
        spine.set_visible(True)
    ax.set_facecolor(facecolor)
    # Add lines to frame
    args = dict(color=grid_linecolor, linestyle="-", linewidth=border_linewidth)
    if jmd_n_len > 0:
        ax.axvline(jmd_n_len, **args)
    if jmd_c_len > 0:
        ax.axvline(x=jmd_n_len + tmd_len, **args)
    return ax


# Main plotting function
def plot_heatmap_(df_feat=None, df_cat=None,
                  col_cat="subcategory", col_val="mean_dif",
                  ax=None, figsize=(8, 8),
                  start=1, tmd_len=20, jmd_n_len=10, jmd_c_len=10,
                  tmd_seq=None, jmd_n_seq=None, jmd_c_seq=None,
                  tmd_color="mediumspringgreen", jmd_color="blue",
                  tmd_seq_color="black", jmd_seq_color="white",
                  seq_size=None,
                  fontsize_tmd_jmd=None, weight_tmd_jmd="normal",
                  fontsize_labels=11,
                  add_xticks_pos=False,
                  grid_linewidth=0.01, grid_linecolor=None,
                  border_linewidth=2,
                  facecolor_dark=False, vmin=None, vmax=None,
                  cmap=None, cmap_n_colors=101,
                  cbar_ax=None, cbar_pct=True, cbar_kws=None,
                  dict_color=None, legend_kws=None,
                  xtick_size=11.0, xtick_width=2.0, xtick_length=5.0):
    """Main function to plot heatmap for feature value per categories/subcategories per position."""
    df_feat = _adjust_df_feat(df_feat=df_feat, col_val=col_val, cbar_pct=cbar_pct)

    # Group arguments
    args_seq = dict(jmd_n_seq=jmd_n_seq, tmd_seq=tmd_seq, jmd_c_seq=jmd_c_seq)
    args_len = dict(tmd_len=tmd_len, jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len)
    args_part_color = dict(tmd_color=tmd_color, jmd_color=jmd_color)
    args_seq_color = dict(tmd_seq_color=tmd_seq_color, jmd_seq_color=jmd_seq_color)
    args_fs = dict(seq_size=seq_size, fontsize_tmd_jmd=fontsize_tmd_jmd)
    args_xtick = dict(xtick_size=xtick_size, xtick_width=xtick_width, xtick_length=xtick_length)

    # Get df positions
    value_type = _get_value_type(col_val=col_val)
    pp = PlotPartPositions(**args_len, start=start)
    df_pos = pp.get_df_pos(df_feat=df_feat.copy(), df_cat=df_cat.copy(),
                           col_cat=col_cat, col_val=col_val,
                           value_type=value_type, normalize=False)
    vmin, vmax = _get_vmin_vmax(df_pos=df_pos, vmin=vmin, vmax=vmax)
    # Get colorbar arguments
    cmap = _get_cmap_heatmap(df_pos=df_pos, cmap=cmap, n_colors=cmap_n_colors, facecolor_dark=facecolor_dark,
                             vmin=vmin, vmax=vmax)
    dict_cbar, cbar_kws = _get_cbar_args_heatmap(df_pos=df_pos, cbar_kws=cbar_kws)

    # Plotting
    pe = PlotElements()
    fig, ax = pe.set_figsize(ax=ax, figsize=figsize, force_set=True)
    ax = _plot_inner_heatmap(df_pos=df_pos, ax=ax, figsize=figsize,
                             start=start, **args_len,
                             grid_linewidth=grid_linewidth, grid_linecolor=grid_linecolor,
                             border_linewidth=border_linewidth,
                             facecolor_dark=facecolor_dark, vmin=vmin, vmax=vmax,
                             cbar_ax=cbar_ax, cmap=cmap, cbar_kws=cbar_kws,
                             x_shift=0.5, **args_xtick)
    # Add colorbar
    _set_cbar_heatmap(ax=ax, vmin=vmin, vmax=vmax,
                      dict_cbar=dict_cbar,
                      cbar_kws=cbar_kws, cbar_pct=cbar_pct,
                      weight="normal", fontsize=fontsize_labels)

    # Add tmd_jmd sequence
    if isinstance(tmd_seq, str):
        ax = pp.add_tmd_jmd_seq(ax=ax, **args_fs,
                                weight_tmd_jmd=weight_tmd_jmd,
                                **args_seq, **args_part_color, **args_seq_color,
                                add_xticks_pos=add_xticks_pos, heatmap=True,
                                x_shift=0.5, xtick_size=xtick_size)
    # Add tmd_jmd bar
    else:
        pp.add_tmd_jmd_bar(ax=ax, **args_part_color)
        pp.add_tmd_jmd_xticks(ax=ax, x_shift=0.5, **args_xtick)
        pp.add_tmd_jmd_text(ax=ax, x_shift=0,
                            fontsize_tmd_jmd=fontsize_tmd_jmd,
                            weight_tmd_jmd=weight_tmd_jmd)

    # Add scale bars
    bar_width = _get_bar_width(fig=fig, len_seq=jmd_n_len+tmd_len+jmd_c_len)
    pe.add_subcat_bars(ax=ax, df_pos=df_pos, df_feat=df_feat,
                       col_cat=col_cat, dict_color=dict_color,
                       bar_width=bar_width, bar_spacing=bar_width*0.75)

    # Add scale legend
    ut.plot_legend_(ax=ax, dict_color=dict_color, **legend_kws)

    # Set current axis to main axis object
    plt.sca(plt.gcf().axes[0])
    ax = plt.gca()
    return ax



"""
This is a script for general CPP Plot check functions.
"""
import time
import pandas as pd
import numpy as np
import warnings

import aaanalysis.utils as ut


# Check for plotting methods
def check_args_xtick(xtick_size=None, xtick_width=None, xtick_length=None):
    """Check if x tick parameters non-negative float"""
    args = dict(accept_none=True, just_int=False, min_val=0)
    ut.check_number_range(name="xtick_size", val=xtick_size, **args)
    ut.check_number_range(name="xtick_width", val=xtick_width, **args)
    ut.check_number_range(name="xtick_length", val=xtick_length, **args)
    args_xtick = dict(xtick_size=xtick_size, xtick_width=xtick_width, xtick_length=xtick_length)
    return args_xtick


def check_args_ytick(ytick_size=None, ytick_width=None, ytick_length=None):
    """Check if y tick parameters non-negative float"""
    args = dict(accept_none=True, just_int=False, min_val=0)
    ut.check_number_range(name="ytick_size", val=ytick_size, **args)
    ut.check_number_range(name="ytick_width", val=ytick_width, **args)
    ut.check_number_range(name="ytick_length", val=ytick_length, **args)
    args_ytick = dict(ytick_size=ytick_size, ytick_width=ytick_width, ytick_length=ytick_length)
    return args_ytick


# Check colors
def check_part_color(tmd_color=None, jmd_color=None):
    """Check if part colors valid"""
    ut.check_color(name="tmd_color", val=tmd_color)
    ut.check_color(name="jmd_color", val=jmd_color)
    args_part_color = dict(tmd_color=tmd_color, jmd_color=jmd_color)
    return args_part_color


def check_seq_color(tmd_seq_color=None, jmd_seq_color=None):
    """Check sequence colors"""
    ut.check_color(name="tmd_seq_color", val=tmd_seq_color, accept_none=True)
    ut.check_color(name="jmd_seq_color", val=jmd_seq_color, accept_none=True)
    args_seq_color = dict(tmd_seq_color=tmd_seq_color, jmd_seq_color=jmd_seq_color)
    return args_seq_color


def check_match_dict_color_df(dict_color=None, df=None, name_df="df_feat"):
    """Check if color dictionary is matching to DataFrame with categories"""
    ut.check_df(name=name_df, df=df, cols_requiered=ut.COL_CAT)
    list_cats = list(sorted(set(df[ut.COL_CAT])))
    if dict_color is None:
        dict_color = ut.DICT_COLOR_CAT
    str_add = f"'dict_color' should be a dictionary with colors for: {list_cats}"
    ut.check_dict_color(val=dict_color, str_add=str_add)
    list_cat_not_in_dict_cat = [x for x in list_cats if x not in dict_color]
    if len(list_cat_not_in_dict_cat) > 0:
        error = f"'dict_color' not complete! Following categories are missing from '{name_df}': {list_cat_not_in_dict_cat}"
        raise ValueError(error)
    for key in list_cats:
        color = dict_color[key]
        ut.check_color(name=key, val=color)
    # Filter colors
    _dict_color = {cat: dict_color[cat] for cat in list_cats}
    return _dict_color



"""
This is the core script for utility functions, folder structure, and constants.

Utility functions are explicitly imported here such that other modules can import them (via ut.).
These comprise options, datatypes, decorators, check functions, internal utility functions,
and backend of external utility functions.
"""
import os
import platform
from functools import lru_cache
import pandas as pd
import seaborn as sns
import numpy as np

# Options
from .config import (options,
                     check_verbose,
                     check_n_jobs,
                     check_random_state)

# Data types
from ._utils.utils_types import (ArrayLike1D,
                                 ArrayLike2D,
                                 VALID_INT_TYPES,
                                 VALID_FLOAT_TYPES,
                                 VALID_INT_FLOAT_TYPES)

# Decorators
from ._utils.decorators import (catch_runtime_warnings,
                                CatchRuntimeWarnings,
                                catch_convergence_warning,
                                ClusteringConvergenceException,
                                catch_invalid_divide_warning,
                                catch_undefined_metric_warning,
                                CatchUndefinedMetricWarning)

# Check functions
from ._utils.check_type import (check_number_range,
                                check_number_val,
                                check_str,
                                check_bool,
                                check_dict,
                                check_tuple,
                                check_list_like)
from ._utils.check_data import (check_X,
                                check_X_unique_samples,
                                check_labels,
                                check_match_X_labels,
                                check_match_X_list_labels,
                                check_match_list_labels_names_datasets,
                                check_array_like,
                                check_superset_subset,
                                check_df)
from ._utils.check_models import (check_mode_class,
                                  check_model_kwargs)
from ._utils.check_plots import (check_fig,
                                 check_ax,
                                 check_figsize,
                                 check_grid_axis,
                                 check_font_weight,
                                 check_fontsize_args,
                                 check_vmin_vmax,
                                 check_lim,
                                 check_dict_xlims,
                                 check_color,
                                 check_list_colors,
                                 check_dict_color,
                                 check_cmap,
                                 check_palette)

# Internal utility functions
from ._utils.utils_output import (print_out,
                                  print_start_progress,
                                  print_progress,
                                  print_end_progress)
from ._utils.utils_plot_elements import (plot_add_bars,
                                         adjust_spine_to_middle,
                                         x_ticks_0,
                                         adjust_tuple_elements)

# External (system-level) utility functions (only backend)
from ._utils.plotting import (plot_gco,
                              plot_get_clist_,
                              plot_legend_)
from ._utils.metrics import (auc_adjusted_,
                             kullback_leibler_divergence_,
                             bic_score_)

# Folder structure
def _folder_path(super_folder, folder_name):
    """Modification of separator (OS-depending)"""
    path = os.path.join(super_folder, folder_name + SEP)
    return path


SEP = "\\" if platform.system() == "Windows" else "/"
FOLDER_PROJECT = os.path.dirname(os.path.abspath(__file__))
FOLDER_DATA = _folder_path(FOLDER_PROJECT, '_data')
URL_DATA = "https://github.com/breimanntools/aaanalysis/tree/master/aaanalysis/data/"


# Constants
FONT_AA = "DejaVu Sans Mono"
STR_AA_GAP = "-"

# Part names
LIST_ALL_PARTS = ["tmd", "tmd_e", "tmd_n", "tmd_c", "jmd_n", "jmd_c", "ext_c", "ext_n",
                  "tmd_jmd", "jmd_n_tmd_n", "tmd_c_jmd_c", "ext_n_tmd_n", "tmd_c_ext_c"]
LIST_PARTS = ["tmd", "jmd_n_tmd_n", "tmd_c_jmd_c"]

# Split names
STR_SEGMENT = "Segment"
STR_PATTERN = "Pattern"
STR_PERIODIC_PATTERN = "PeriodicPattern"
LIST_SPLIT_TYPES = [STR_SEGMENT, STR_PATTERN, STR_PERIODIC_PATTERN]
SPLIT_DESCRIPTION = f"\n a) {STR_SEGMENT}(i-th,n_split)" \
                    f"\n b) {STR_PATTERN}(N/C,p1,p2,...,pn)" \
                    f"\n c) {STR_PERIODIC_PATTERN}(N/C,i+step1/step2,start)" \
                    f"\nwith i-th<=n_split, and p1<p2<...<pn," \
                    f" where all numbers should be non-negative integers, and N/C means 'N' or 'C'."

# Scale dataset names
STR_SCALES = "scales"   # Min-max normalized scales (from AAontology)
STR_SCALES_RAW = "scales_raw"   # Raw scales (from AAontology)
STR_SCALES_PC = "scales_pc"     # AAclust pc-based scales (pc: principal component)
STR_SCALE_CAT = "scales_cat"  # AAontology
STR_TOP60 = "top60"    # AAclustTop60
STR_TOP60_EVAL = "top60_eval"  # AAclustTop60 evaluation
NAMES_SCALE_SETS = [STR_SCALES, STR_SCALES_RAW, STR_SCALE_CAT,
                    STR_SCALES_PC, STR_TOP60, STR_TOP60_EVAL]


# Column names for primary df
# df_seq
COL_ENTRY = "entry"     # ACC, protein entry, uniprot id
COL_NAME = "name"       # Entry name, Protein name, Uniprot Name
COL_LABEL = "label"
COL_SEQ = "sequence"
COL_JMD_N = "jmd_n"
COL_TMD = "tmd"
COL_JMD_C = "jmd_c"
COL_TMD_START = "tmd_start"
COL_TMD_STOP = "tmd_stop"
COLS_SEQ_INFO = [COL_ENTRY, COL_SEQ, COL_LABEL]
COLS_SEQ_POS = [COL_SEQ, COL_TMD_START, COL_TMD_STOP]
COLS_SEQ_PARTS = [COL_JMD_N, COL_TMD, COL_JMD_C]
COLS_SEQ_TMD = [COL_SEQ, COL_TMD]

# df_part

# df_scales
# Column for df_cat (as defined in AAontology, retrieved by aa.load_scales(name="scale_cat"))
COL_SCALE_ID = "scale_id"
COL_CAT = "category"
COL_SUBCAT = "subcategory"
COL_SCALE_NAME = "scale_name"
COL_SCALE_DES = "scale_description"

# Columns for df_feat
COL_FEATURE = "feature"
# COL_CAT, COL_SUBCAT, COL_SCALE_NAME, COL_SCALE_DES
COL_ABS_AUC = "abs_auc"
COL_ABS_MEAN_DIF = "abs_mean_dif"
COL_MEAN_DIF = "mean_dif"
COL_STD_TEST = "std_test"
COL_STD_REF = "std_ref"
COL_PVAL_MW = "p_val_mann_whitney"
COL_PVAL_FDR = "p_val_fdr_bh"
COL_POSITION = "positions"
COL_AA_TEST = "amino_acids_test"
COL_AA_REF = "amino_acids_ref"

# Columns for df_feat after processing with explainable AI methods
COL_FEAT_IMPORT = "feat_importance"
COL_FEAT_IMPORT_STD = "feat_importance_std"
COL_FEAT_IMPACT = "feat_impact"
COL_FEAT_IMPACT_STD = "feat_impact_std"

COLS_FEAT_SCALES = [COL_CAT, COL_SUBCAT, COL_SCALE_NAME]
COLS_FEAT_STAT = [COL_ABS_AUC, COL_ABS_MEAN_DIF, COL_MEAN_DIF, COL_STD_TEST, COL_STD_REF]
COLS_FEAT_WEIGHT = [COL_FEAT_IMPORT, COL_FEAT_IMPORT_STD, COL_FEAT_IMPACT]
DICT_VALUE_TYPE = {COL_ABS_AUC: "mean",
                   COL_ABS_MEAN_DIF: "mean",
                   COL_MEAN_DIF: "mean",
                   COL_STD_TEST: "mean",
                   COL_STD_REF: "mean",
                   COL_FEAT_IMPORT: "sum",
                   COL_FEAT_IMPORT_STD: "mean",
                   COL_FEAT_IMPACT: "sum"}


# Columns of df_eval
# AAclust (evaluation)
METRIC_CORRELATION = "correlation"
LIST_METRICS = [METRIC_CORRELATION, "manhattan",  "euclidean", "cosine"]
STR_UNCLASSIFIED = "Unclassified"
COL_N_CLUST = "n_clusters"
COL_BIC = "BIC"
COL_CH = "CH"
COL_SC = "SC"
COL_RANK = "rank"
COLS_EVAL_AACLUST = [COL_N_CLUST, COL_BIC, COL_CH, COL_SC]

# CPP evaluation
COL_N_FEAT = "n_features"
COL_AVG_ABS_AUC = "avg_ABS_AUC"
COL_RANGE_ABS_AUC = "range_ABS_AUC"   # Quintiles (min, 25%, median, 75%, max)
COL_AVG_MEAN_DIF = "avg_MEAN_DIF"
COL_AVG_STD_TEST = "avg_STD_TEST"
# COL_N_CLUST = "n_clusters"
COL_AVG_N_FEAT_PER_CLUST = "avg_n_feat_per_clust"
COL_STD_N_FEAT_PER_CLUST = "std_n_feat_per_clust"
COL_EVAL_CPP = [COL_N_FEAT,
                COL_AVG_ABS_AUC, COL_AVG_MEAN_DIF, COL_AVG_STD_TEST,
                COL_N_CLUST, COL_AVG_N_FEAT_PER_CLUST, COL_STD_N_FEAT_PER_CLUST]

# dPULearn (evaluation)
COL_SELECTION_VIA = "selection_via"
COL_N_REL_NEG = "n_rel_neg"
COL_AVG_STD = "avg_STD"
COL_AVG_IQR = "avg_IQR"
COL_AVG_ABS_AUC_POS = "avg_abs_AUC_pos"
COL_AVG_KLD_POS = "avg_KLD_pos"
COL_AVG_ABS_AUC_UNL = "avg_abs_AUC_unl"
COL_AVG_KLD_UNL = "avg_KLD_unl"
COL_AVG_ABS_AUC_NEG = "avg_abs_AUC_neg"
COL_AVG_KLD_NEG = "avg_KLD_neg"
COLS_EVAL_DPULEARN_SIMILARITY = [COL_AVG_STD, COL_AVG_IQR]
COLS_EVAL_DPULEARN_DISSIMILARITY = [COL_AVG_ABS_AUC_POS, COL_AVG_KLD_POS,
                                    COL_AVG_ABS_AUC_UNL, COL_AVG_KLD_UNL,
                                    COL_AVG_ABS_AUC_NEG, COL_AVG_KLD_NEG]
COLS_EVAL_DPULEARN = [COL_N_REL_NEG] + COLS_EVAL_DPULEARN_SIMILARITY + COLS_EVAL_DPULEARN_DISSIMILARITY

# Labels
LABEL_FEAT_VAL = "Feature value"
LABEL_HIST_COUNT = "Number of proteins"
LABEL_HIST_DEN = "Relative density"

LABEL_FEAT_NUMBER = "Number of features\n(per residue position)"
LABEL_FEAT_IMPORT_CUM = "Cumulative feature importance\n(normalized) [%]"
LABEL_FEAT_IMPACT_CUM = "Cumulative feature impact\n(normalized) [%]"
LABEL_CBAR_FEAT_IMPACT_CUM = "Cumulative feature impact"
LABEL_FEAT_POS = ("                  Feature                  \n"
                  "Scale (subcategory)  +  Positions                 ")

_LABEL_FEAT_POS = ("      Feature     \n"
                  "       Scale  +  Positions")

LABEL_FEAT_IMPORT = "Importance [%]"
LABEL_FEAT_IMPACT = "Impact [%]"
LABEL_FEAT_RANKING = "Feature ranking"
LABEL_SCALE_CAT = "Scale category"
LABEL_MEAN_DIF = "Mean difference"

# Standard colors
COLOR_SHAP_POS = '#FF0D57'  # (255, 13, 87)
COLOR_SHAP_NEG = '#1E88E5'  # (30, 136, 229)
COLOR_FEAT_POS = '#9D2B39'  # (157, 43, 57) Mean difference
COLOR_FEAT_NEG = '#326599'  # (50, 101, 133) Mean difference
COLOR_FEAT_IMP = '#7F7F7F'  # (127, 127, 127) feature importance
COLOR_TMD = '#00FA9A'       # (0, 250, 154)
COLOR_JMD = '#0000FF'       # (0, 0, 255)
COLOR_POS =  "#389d2b"    # (56, 157, 43)
COLOR_UNL = "tab:gray"
COLOR_NEG = "#ad4570"   # (173,69,112)
COLOR_REL_NEG = "#ad9745" # (173, 151, 69)

DICT_COLOR = {"SHAP_POS": COLOR_SHAP_POS,
              "SHAP_NEG": COLOR_SHAP_NEG,
              "FEAT_POS": COLOR_FEAT_POS,
              "FEAT_NEG": COLOR_FEAT_NEG,
              "FEAT_IMP": COLOR_FEAT_IMP,
              "TMD": COLOR_TMD,
              "JMD": COLOR_JMD,
              "SAMPLES_POS": COLOR_POS,
              "SAMPLES_UNL": COLOR_UNL,
              "SAMPLES_NEG": COLOR_NEG,
              "SAMPLES_REL_NEG": COLOR_REL_NEG
              }

DICT_COLOR_CAT = {"ASA/Volume": "tab:blue",
                  "Composition": "tab:orange",
                  "Conformation": "tab:green",
                  "Energy": "tab:red",
                  "Others": "tab:gray",
                  "Polarity": "gold",
                  "Shape": "tab:cyan",
                  "Structure-Activity": "tab:brown"}

LIST_CAT = ['ASA/Volume', 'Composition', 'Conformation', 'Energy',
            'Others', 'Polarity', 'Shape', 'Structure-Activity']


# Parameter options for cmaps and color dicts
STR_CMAP_CPP = "CPP"
STR_CMAP_SHAP = "SHAP"
STR_DICT_COLOR = "DICT_COLOR"
STR_DICT_CAT = "DICT_CAT"



# I Helper functions
def _retrieve_string_starting_at_end(seq, start=None, end=None):
    """Reverse_string_start_end"""
    def reverse_string(s):
        return s[::-1]
    reversed_seq = reverse_string(seq)
    reversed_seq_part = reversed_seq[start:end]
    seq = reverse_string(reversed_seq_part)
    return seq


def get_dict_part_seq(tmd=None, jmd_n=None, jmd_c=None):
    """Get dictionary for part to sequence"""
    # Length of extending part (starting from C and N terminal part of TMD)
    ext_len = options["ext_len"]
    tmd_n = tmd[0:round(len(tmd) / 2)]
    tmd_c = tmd[round(len(tmd) / 2):]
    # Historical feature parts (can be set via aa.options["ext_len"] = 4
    ext_n = _retrieve_string_starting_at_end(jmd_n, start=0, end=ext_len)  # helix_stop motif for TMDs
    ext_c = jmd_c[0:ext_len]  # anchor for TMDs
    tmd_e = ext_n + tmd + ext_c
    part_seq_dict = {'tmd': tmd, 'tmd_e': tmd_e,
                     'tmd_n': tmd_n, 'tmd_c': tmd_c,
                     'jmd_n': jmd_n, 'jmd_c': jmd_c,
                     'ext_n': ext_n, 'ext_c': ext_c,
                     'tmd_jmd': jmd_n + tmd + jmd_c,
                     'jmd_n_tmd_n': jmd_n + tmd_n, 'tmd_c_jmd_c': tmd_c + jmd_c,
                     'ext_n_tmd_n': ext_n + tmd_n, 'tmd_c_ext_c': tmd_c + ext_c}
    return part_seq_dict





# II Main functions
# Caching for data loading for better performance (data loaded ones)
@lru_cache(maxsize=None)
def read_excel_cached(name, index_col=None):
    """Load cached dataframe to save loading time"""
    df = pd.read_excel(name, index_col=index_col)
    return df.copy()


@lru_cache(maxsize=None)
def read_csv_cached(name, sep=None):
    """Load cached dataframe to save loading time"""
    df = pd.read_csv(name, sep=sep)
    return df.copy()


def load_default_scales(scale_cat=False):
    """Load default scales sets or categories. Copy is always returned to maintain data integrity."""
    if scale_cat:
        if options["df_cat"] is None:
            df_cat = read_excel_cached(FOLDER_DATA + f"{STR_SCALE_CAT}.xlsx")
            options["df_cat"] = df_cat
            return df_cat.copy()
        else:
            return options["df_cat"].copy()
    else:
        if options["df_scales"] is None:
            df_scales = read_excel_cached(FOLDER_DATA + f"{STR_SCALES}.xlsx", index_col=0)
            df_scales = df_scales.astype(float)
            options["df_scales"] = df_scales
            return df_scales.copy()
        else:
            return options["df_scales"].copy()


# Adjust df_eval
def add_names_to_df_eval(df_eval=None, names=None):
    """Add names column to df_eval"""
    if names is None:
        n_datasets = len(df_eval)
        names = [f"Set {i}" for i in range(1, n_datasets + 1)]
    df_eval.insert(0, COL_NAME, names)
    return df_eval


# Plotting utilities
def get_color_dif(mean_dif=0):
    return COLOR_FEAT_NEG if mean_dif < 0 else COLOR_FEAT_POS


# DEV: Exceptionally placed here due to dependency on constants
def plot_get_cdict_(name=STR_DICT_COLOR):
    """Return DICT_COLOR or DICT_COLOR_CAT"""
    if name == STR_DICT_COLOR:
        return DICT_COLOR
    elif name == STR_DICT_CAT:
        return DICT_COLOR_CAT
    else:
        raise ValueError(f"'name' must be '{STR_DICT_COLOR}' or '{STR_DICT_CAT}'")


def _get_diverging_cmap(cmap="ReBu_r", n_colors=101, facecolor_dark=False, only_pos=False, only_neg=False):
    """Generate a diverging colormap based on the provided cmap."""
    n = min(int(np.floor(1 + n_colors/20)), 5)
    color_0 = [(0, 0, 0)] if facecolor_dark else [(1, 1, 1)]
    if only_neg:
        cmap = sns.color_palette(palette=cmap, n_colors=(n_colors * 2) + n)
        cmap = cmap[0:n_colors-1] + color_0
    elif only_pos:
        cmap = sns.color_palette(palette=cmap, n_colors=(n_colors * 2) + n)
        cmap = color_0 + cmap[-n_colors+1:]
    else:
        n_cmap = n_colors + n * 2
        n_colors_half = int(np.floor(n_colors / 2))
        n_sub = (n_colors + 1) % 2
        n_cmap_half = int(np.floor(n_cmap / 2))
        cmap = sns.color_palette(cmap, n_colors=n_cmap)
        cmap_low, cmap_high = cmap[:n_cmap_half - n], cmap[n + n_cmap_half:]
        cmap = cmap_low[0:n_colors_half-n_sub] + color_0 + cmap_high[-n_colors_half:]
    return cmap


def _get_shap_cmap(n_colors=101, facecolor_dark=True, only_pos=False, only_neg=False):
    """Generate a diverging colormap for feature values."""
    n = min((int(np.floor(1 + n_colors/5)), 20))
    color_0 = [(0, 0, 0)] if facecolor_dark else [(1, 1, 1)]

    if only_neg:
        cmap = sns.light_palette(COLOR_SHAP_NEG, input="hex", reverse=True, n_colors=n_colors + n)
        cmap = cmap[0:n_colors-1] + color_0
    elif only_pos:
        cmap = sns.light_palette(COLOR_SHAP_POS, input="hex", n_colors=n_colors + n)
        cmap = color_0 + cmap[-n_colors+1:]
    else:
        n_colors_half = int(np.floor(n_colors / 2))
        n_sub = (n_colors + 1) % 2
        n_cmap_half = n_colors_half + n
        cmap_low = sns.light_palette(COLOR_SHAP_NEG, input="hex", reverse=True, n_colors=n_cmap_half)
        cmap_high = sns.light_palette(COLOR_SHAP_POS, input="hex", n_colors=n_cmap_half)
        cmap = cmap_low[0:n_colors_half-n_sub] + color_0 + cmap_high[-n_colors_half:]
    return cmap


def plot_get_cmap_(cmap="CPP", n_colors=101, facecolor_dark=None, only_pos=False, only_neg=False):
    """Get colormap for CPP or CPP-SHAP plots"""
    args = dict(n_colors=n_colors, facecolor_dark=facecolor_dark,
                only_neg=only_neg, only_pos=only_pos)
    if cmap == STR_CMAP_CPP:
        cmap = _get_diverging_cmap("RdBu_r", **args)
    elif cmap == STR_CMAP_SHAP:
        cmap =  _get_shap_cmap(**args)
    else:
        cmap = _get_diverging_cmap(cmap=cmap, **args)
    return cmap


# Check df_seq
def check_df_seq(df_seq=None, accept_none=False):
    """Check columns from df_seq is valid regarding four distinct formats, differentiated by their respective columns:
        a) Position-based format: ['sequence', 'tmd_start', 'tmd_stop']
        b) Part-based format: ['jmd_n', 'tmd', 'jmd_c']
        c) Sequence-based format: ['sequence']
        d) Sequence-TMD-based format: ['sequence', 'tmd']
    """
    if df_seq is None:
        if accept_none:
            return None
        else:
            raise ValueError("'df_seq' should not be None")
    check_df(name="df_seq", df=df_seq, cols_requiered=[COL_ENTRY])
    pos_based = set(COLS_SEQ_POS).issubset(set(df_seq))
    part_based = set(COLS_SEQ_PARTS).issubset(set(df_seq))
    seq_based = COL_SEQ in list(df_seq)
    seq_tmd_based = set(COLS_SEQ_TMD).issubset(set(df_seq))
    if not (seq_based or pos_based or part_based or seq_tmd_based):
        raise ValueError(f"'df_seq' should contain one of the following sets of columns:"
                         f" a) {COLS_SEQ_POS} (Position-based format)"
                         f" b) {COLS_SEQ_PARTS} (Part-based format)"
                         f" c) {[COL_SEQ]} (Sequence-based format)"
                         f" d) {COLS_SEQ_TMD} (Sequence-TMD-based format)")
    # Check tmd_start & tmd_stop columns
    if "start" in list(df_seq):
        raise ValueError(f"'df_seq' should not contain 'tmd_start' in columns. Change column to '{COL_TMD_START}'.")
    if "stop" in list(df_seq):
        raise ValueError(f"'df_seq' should not contain 'tmd_stop' in columns. Change column to '{COL_TMD_STOP}'.")
    # Check if different formats are valid
    if pos_based:
        for entry, tmd_start, tmd_stop in zip(df_seq[COL_ENTRY], df_seq[COL_TMD_START], df_seq[COL_TMD_STOP]):
            check_number_range(name=f"'tmd_start'={tmd_start} (entry: '{entry}')", val=tmd_start, just_int=True)
            check_number_range(name=f"'tmd_stop'={tmd_stop} (entry: '{entry}')", val=tmd_stop, just_int=True)
    if part_based:
        for col in COLS_SEQ_PARTS:
            if not all(isinstance(x, str) for x in df_seq[col]):
                raise ValueError(f"'{col}' should only contain strings.")
    if seq_based:
        if not all(isinstance(x, str) for x in df_seq[COL_SEQ]):
            raise ValueError(f"'{COL_SEQ}' should only contain strings.")
    if seq_tmd_based:
        if any([tmd not in seq for tmd, seq in zip(df_seq[COL_TMD], df_seq[COL_SEQ])]):
            raise ValueError(f"Parts in '{COL_TMD}' should be contained in '{COL_SEQ}'")
    # Check matching of parts with 'sequence', 'tmd_start', and 'tmd_stop'
    if part_based and pos_based:
        for i, row in df_seq.iterrows():
            entry = row[COL_ENTRY]
            jmd_n, tmd, jmd_c = row[COL_JMD_N], row[COL_TMD], row[COL_JMD_C]
            tmd_jmd = jmd_n + tmd + jmd_c
            seq, start, stop = row[COL_SEQ], row[COL_TMD_START], row[COL_TMD_STOP]
            if tmd_jmd not in seq:
                raise ValueError(f"For '{entry}' entry, '{COL_JMD_N}', '{COL_TMD}', and '{COL_JMD_C}' "
                                 f"do not match with '{COL_SEQ}'")
            if seq[start-1:stop] != tmd:
                raise ValueError(f"For '{entry}' entry, '{COL_TMD_START}' and '{COL_TMD_STOP}' "
                                 f"do not match with '{COL_TMD}'")


# Check parts
def check_list_parts(list_parts=None, return_default=True, all_parts=False, accept_none=False):
    """Check if parts from list_parts are columns of df_seq"""
    try:
        list_parts = check_list_like(name="list_parts", val=list_parts, accept_none=True,
                                     accept_str=True, convert=True)
    except ValueError:
        raise ValueError(f"'list_parts' must be list with selection of following parts: {LIST_ALL_PARTS}")
    if list_parts is None:
        if return_default:
            list_parts = LIST_ALL_PARTS if all_parts else LIST_PARTS
            # Remove ext parts if ext is 0
            if options["ext_len"] == 0:
                list_parts = [l for l in list_parts if "ext" not in l and l != "tmd_e"]
            return list_parts
        elif accept_none:
            return  # skip further checks
        else:
            raise ValueError(f"'list_parts' must be list with selection of following parts: {LIST_ALL_PARTS}")
    # Check if empty list
    if len(list_parts) == 0:
        raise ValueError(f"'list_parts' should not be empty list.")
    # Check for invalid parts
    wrong_parts = [x for x in list_parts if x not in LIST_ALL_PARTS]
    if len(wrong_parts) > 0:
        str_part = "part" if len(wrong_parts) == 1 else "parts"
        error = f"{wrong_parts} not valid {str_part}.\n  Select from following parts: {LIST_ALL_PARTS}"
        raise ValueError(error)
    return list_parts


# Check features
def _check_part(part=None, feature=None, list_parts=None):
    """Check if feature PART is valid"""
    list_parts = check_list_like(name="list_parts", val=list_parts, accept_str=True)
    list_parts = [x.lower() for x in list_parts]
    part = part.replace(" ", "")  # remove whitespace
    error = f"Wrong 'PART' for '{feature}'. Features should be 'PART-SPLIT-SCALE', with parts from: {list_parts}."
    if part.lower() in LIST_ALL_PARTS:
        error += f"\n Or include '{part.lower()}' in parts or set 'ext_len' option > 0."
    if part.lower() not in list_parts:
        raise ValueError(error)


def _check_split(split=None, feature=None):
    """Check feature SPLIT has valid form"""
    split = split.replace(" ", "")  # remove whitespace
    error = f"Wrong 'SPLIT' for '{feature}'. Features should be 'PART-SPLIT-SCALE', with splits denoted as follows: {SPLIT_DESCRIPTION}"
    # Check Segment
    if STR_SEGMENT in split:
        try:
            i_th, n_split = [int(x) for x in split.split("(")[1].replace(")", "").split(",")]
        except:
            error += "\n Error: Wrong Segment."
            raise ValueError(error)
        # Check if values non-negative integers
        for name, val in zip(["i_th", "n_split"], [i_th, n_split]):
            check_number_range(name=name, val=val, just_int=True)
        # Check if i-th and n_split are valid
        if i_th > n_split:
            error += "\n Error: i-th segment should be smaller than 'n_splits'."
            raise ValueError(error)
    # Check PeriodicPattern
    elif STR_PERIODIC_PATTERN in split:
        try:
            start = split.split("i+")[1].replace(")", "").split(",")
            step1, step2 = [int(x) for x in start.pop(0).split("/")]
            start = int(start[0])
        except:
            error += "\n Error: Wrong PeriodicPattern."
            raise ValueError(error)
        # Check if values non-negative integers
        for name, val in zip(["start", "step1", "step2"], [start, step1, step2]):
            check_number_range(name=name, val=val, min_val=1, just_int=True)
        # Check if terminus valid
        terminus = split.split("i+")[0].split("(")[1].replace(",", "")
        if terminus not in ["N", "C"]:
            error += "\n Error: Terminus should be 'N' or 'C'."
            raise ValueError(error)
    # Check pattern (must be after PeriodicPattern due to string matching)
    elif STR_PATTERN in split:
        try:
            list_pos = split.split("(")[1].replace(")", "").split(",")
            terminus = list_pos.pop(0)
            list_pos = [int(x) for x in list_pos]
        except:
            error += "\n Error: Wrong Pattern."
            raise ValueError(error)
        # Check if contain at least one position
        if len(list_pos) < 1:
            error += "\n Error: Steps should contain at least 1 element."
            raise ValueError(error)
        # Check if values non-negative integers
        for val in list_pos:
            name = "pos" + str(val)
            check_number_range(name=name, val=val, min_val=1, just_int=True)
        # Check if terminus valid
        if terminus not in ["N", "C"]:
            error += "\n Error: Terminus should be 'N' or 'C'."
            raise ValueError(error)
        # Check if arguments are in order
        if not sorted(list_pos) == list_pos:
            error += "\n Error: Positions are in wrong order."
            raise ValueError(error)
    else:
        raise ValueError(error)


def _check_scale(scale=None, feature=None, list_scales=None):
    """Check if scale in feature is valid"""
    list_scales = check_list_like(name="list_scales", val=list_scales, accept_none=True, accept_str=True)
    if list_scales is not None:
        error = f"Wrong 'SCALE' for '{feature}'. Features should be 'PART-SPLIT-SCALE', with scales from: {LIST_ALL_PARTS}"
        if scale not in list_scales:
            raise ValueError(error)


def check_features(features=None, list_parts=None, list_scales=None):
    """Check if feature names are valid for list of parts  and df_scales"""
    features = check_list_like(name="features", val=features, accept_none=False, accept_str=True, convert=True)
    list_parts = check_list_parts(list_parts=list_parts, all_parts=True, return_default=True)
    # Check elements of features list
    features_wrong_n_components = [x for x in features if type(x) is not str or len(x.split("-")) != 3]
    if len(features_wrong_n_components) > 0:
        error = (f"Following elements from 'features' are not valid: {features_wrong_n_components}"
                 f"\n  Form of features should be 'PART-SPLIT-SCALE'")
        raise ValueError(error)
    # Check part, split, and scale
    if len(features) == 0:
        raise ValueError("'features' should not be empty.")
    for feature in features:
        part, split, scale = feature.split("-")
        _check_part(part=part, feature=feature, list_parts=list_parts)
        _check_split(split=split, feature=feature)
        # Scales is only checked if list_scales is provided
        _check_scale(scale=scale, feature=feature, list_scales=list_scales)
    return features


def check_df_feat(df_feat=None, df_cat=None, list_parts=None, shap_plot=None,
                  cols_requiered=None, cols_nan_check=None):
    """Check if df not empty pd.DataFrame"""
    # Check df
    cols_feat = [COL_FEATURE] + COLS_FEAT_SCALES + COLS_FEAT_STAT
    if cols_requiered is not None:
        cols_feat += [cols_requiered] if type(cols_requiered) is str else cols_requiered
    check_df(df=df_feat, name="df_feat", cols_requiered=cols_feat, cols_nan_check=cols_nan_check)
    if len(df_feat) == 0 or len(list(df_feat)) == 0:
        raise ValueError("'df_feat' should be not empty")
    duplicated_columns = df_feat.columns[df_feat.columns.duplicated()]
    if len(duplicated_columns) > 0:
        raise ValueError(f"'df_feat' should not contain duplicated columns: {duplicated_columns}")
    # Check features
    features = list(df_feat[COL_FEATURE]) #.values.tolist()
    check_features(features=features, list_parts=list_parts)
    # Check if df_feat matches df_cat
    if df_cat is not None:
        scales = set([x.split("-")[2] for x in features])
        list_scales = list(df_cat[COL_SCALE_ID])
        missing_scales = [x for x in scales if x not in list_scales]
        if len(missing_scales) > 0:
            raise ValueError(f"Following scales occur in 'df_feat' but not in 'df_cat': {missing_scales}")
    # Check if feat_importance or feat_impact column is in df_feat
    if shap_plot is not None:
        if shap_plot:
            col_feat_impact = [x for x in list(df_feat) if COL_FEAT_IMPACT in x]
            if len(col_feat_impact) == 0:
                raise ValueError(f"If 'shap_plot' is True, At least on '{COL_FEAT_IMPACT}' column must be "
                                 f"in 'df_feat' columns: {list(df_feat)}")
        else:
            if COL_FEAT_IMPORT not in list(df_feat):
                raise ValueError(f"If 'shap_plot' is False, '{COL_FEAT_IMPORT}' must be in 'df_feat' columns: {list(df_feat)}")
    df_feat = df_feat.reset_index(drop=True).copy()
    return df_feat



from ._plot_get_clist import plot_get_clist
from ._plot_get_cmap import plot_get_cmap
from ._plot_get_cdict import plot_get_cdict
from ._plot_settings import plot_settings
from ._plot_gcfs import plot_gcfs
from ._plot_legend import plot_legend
from ._display_df import display_df


__all__ = [
    "plot_get_clist",
    "plot_get_cmap",
    "plot_get_cdict",
    "plot_settings",
    "plot_legend",
    "plot_gcfs",
    "display_df",
]



"""
This is a script for the backend of the AAclust.fit() method.

The fit functions performs the AAclust algorithm consisting of three steps:

1. Estimate lower bound for n_clusters
2. Optimization of n_clusters
3. Merge clusters

"""
import pandas as pd
import numpy as np
from collections import OrderedDict
from sklearn.metrics.pairwise import pairwise_distances

import aaanalysis.utils as ut
from ._utils_aaclust import _cluster_center

# I Helper Functions
def min_cor_center(X):
    """Get minimum for correlation of all columns with cluster center, defined as the mean values
    for each amino acid over all scales."""
    # Create new matrix including cluster center
    center_X = np.concatenate([_cluster_center(X), X], axis=0)
    # Get minimum correlation with mean values
    min_cor = np.corrcoef(center_X)[0, ].min()
    return min_cor


def min_cor_all(X):
    """Get minimum for pair-wise correlation of all columns in given matrix."""
    # Get minimum correlations minimum/ maximum distance for pair-wise comparisons
    min_cor = np.corrcoef(X).min()
    return min_cor


def get_min_cor(X, labels=None, on_center=True):
    """Compute minimum pair-wise correlation or correlation with cluster center for each cluster label
    and return minimum of obtained cluster minimums."""
    f = min_cor_center if on_center else min_cor_all
    if labels is None:
        return f(X)
    # Minimum correlations for each cluster (with center or all scales)
    unique_labels = list(OrderedDict.fromkeys(labels))
    list_masks = [[i == label for i in labels] for label in unique_labels]
    list_min_cor = [f(X[mask]) for mask in list_masks]
    # Minimum for all clusters
    min_cor = min(list_min_cor)
    return min_cor


def get_max_dist(X, on_center=True, metric="euclidean"):
    """Get maximum distance on center or all scales"""
    # Maximum distance for cluster
    if on_center:
        # Create new matrix including cluster center
        center_X = np.concatenate([_cluster_center(X), X], axis=0)
        # Get maximum distance with mean values
        max_dist = pairwise_distances(center_X, metric=metric)[0, ].max()
    else:
        # Get maximum distance for pair-wise comparisons
        max_dist = pairwise_distances(X, metric=metric).max()
    return max_dist


# II Main Functions
# 1. Step (Estimation of n clusters)
@ut.catch_convergence_warning()
def _estimate_lower_bound_n_clusters(X, model=None, model_kwargs=None, min_th=0.3, on_center=True):
    """
    Estimate the lower bound of the number of clusters (k).

    This function estimates the lower bound of the number of clusters by testing a range
    between 10% and 90% of all observations, incrementing in 10% steps.
    """
    f = lambda c: get_min_cor(X, labels=model(n_clusters=c, **model_kwargs).fit(X).labels_, on_center=on_center)
    # Create range between 10% and 90% of all scales (10% steps) as long as minimum correlation is lower than threshold
    n_samples, n_features = X.shape
    nclust_mincor = [(1, f(1))]
    step_number = 40
    for i in range(1, step_number, 1):
        n_clusters = max(1, int(n_samples*i/step_number))    # n cluster in 2.5% steps
        min_cor = f(n_clusters)
        if min_cor < min_th:   # Save only lower bounds
            nclust_mincor.append((n_clusters, min_cor))
        else:
            break
    # Select second highest lower bound (highest lower bound is faster but might surpass true bound)
    nclust_mincor.sort(key=lambda x: x[0], reverse=True)
    n_clusters = nclust_mincor[1][0] if len(nclust_mincor) > 1 else nclust_mincor[0][0]  # Otherwise, only existing one
    return n_clusters


@ut.catch_runtime_warnings()
def estimate_lower_bound_n_clusters(X, model=None, model_kwargs=None, min_th=0.6, on_center=True):
    """Wrapper for _estimate_lower_bound_n_clusters to catch convergence warnings"""
    try:
        n_clusters = _estimate_lower_bound_n_clusters(X, model=model, model_kwargs=model_kwargs,
                                                      min_th=min_th, on_center=on_center)
    except ut.ClusteringConvergenceException as e:
        n_clusters = e.distinct_clusters
    return n_clusters


# 2. Step (Optimization of n clusters)
@ut.catch_convergence_warning()
def _optimize_n_clusters(X, model=None, model_kwargs=None, n_clusters=None, min_th=0.3, on_center=True):
    """
    Optimize the number of clusters using a recursive algorithm.

    This function performs clustering in a recursive manner (through a while loop) to ensure
    that the minimum within-cluster correlation is achieved for all clusters. It is an efficiency
    optimized version of a step-wise algorithm where the `n_clusters` is incrementally increased
    until a stop condition is met.
    """
    n_samples, n_features = X.shape
    f = lambda c: get_min_cor(X, labels=model(n_clusters=c, **model_kwargs).fit(X).labels_, on_center=on_center)
    min_cor = f(n_clusters)
    # Recursive optimization of n_clusters via step-wise increase starting from lower bound
    step = max(1, min(int(n_samples/10), 5))    # Step size between 1 and 5
    while min_cor < min_th and n_clusters < n_samples:    # Stop condition of clustering
        n_clusters = min(n_clusters+step, n_samples) # Maximum of n_samples is allowed
        min_cor = f(n_clusters)
        # Exceeding of threshold -> Conservative adjustment of clustering parameters to meet true optimum
        if min_cor >= min_th and step != 1:
            n_clusters = max(1, n_clusters - step * 2)
            step = 1
            min_cor = f(n_clusters)
    return n_clusters


@ut.catch_runtime_warnings()
def optimize_n_clusters(X, model=None, model_kwargs=None, n_clusters=None, min_th=0.5, on_center=True):
    """Wrapper for _optimize_n_clusters to catch convergence warnings"""
    try:
        n_clusters = _optimize_n_clusters(X, model=model, model_kwargs=model_kwargs,
                                          n_clusters=n_clusters, min_th=min_th, on_center=on_center)
    except ut.ClusteringConvergenceException as e:
        n_clusters = e.distinct_clusters
    return n_clusters


# 3. Step (Merging)
def _get_min_cor_cluster(X, labels=None, label_cluster=None, on_center=True):
    """Get min_cor for single cluster"""
    mask = [l == label_cluster for l in labels]
    min_cor = get_min_cor(X[mask], on_center=on_center)
    return min_cor


def _get_quality_measure(X, metric=None, labels=None, label_cluster=None, on_center=True):
    """Get quality measure single cluster given by feature matrix X, labels, and label of cluster"""
    mask = [l == label_cluster for l in labels]
    if metric == ut.METRIC_CORRELATION:
        return get_min_cor(X[mask], on_center=on_center)
    else:
        return get_max_dist(X[mask], on_center=on_center, metric=metric)


def _get_best_cluster(dict_clust_qm=None, metric=None):
    """Get cluster with the best quality measure: either highest minimum Pearson correlation
    or lowest distance measure"""
    if metric == ut.METRIC_CORRELATION:
        return max(dict_clust_qm, key=dict_clust_qm.get)
    else:
        return min(dict_clust_qm, key=dict_clust_qm.get)


def merge_clusters(X, n_max=5, labels=None, min_th=0.5, on_center=True, metric="correlation"):
    """
    Merge small clusters into other clusters optimizing a given quality measure.

    This function merges clusters with sizes less than or equal to `n_max` into other clusters
    based on a specified quality measure (Pearson correlation or a distance metric).
    Merging is conducted only if the new assignment meets a minimum within-cluster Pearson
    correlation threshold defined by `min_th`.
    """
    unique_labels = list(OrderedDict.fromkeys(labels))
    for n in range(1, n_max):
        s_clusters = [x for x in unique_labels if labels.count(x) == n]   # Smallest clusters
        b_clusters = [x for x in unique_labels if labels.count(x) > n]    # Bigger clusters (all others)
        # Assign scales from smaller clusters to cluster by optimizing for quality measure
        for s_clust in s_clusters:
            dict_clust_qm = {}  # Cluster to quality measure
            for b_clust in b_clusters:
                labels_ = [x if x != s_clust else b_clust for x in labels]
                args = dict(labels=labels_, label_cluster=b_clust, on_center=on_center)
                min_cor = _get_min_cor_cluster(X, **args)
                if min_cor >= min_th:
                    dict_clust_qm[b_clust] = _get_quality_measure(X, **args, metric=metric)
            if len(dict_clust_qm) > 0:
                b_clust_best = _get_best_cluster(dict_clust_qm=dict_clust_qm, metric=metric)
                labels = [x if x != s_clust else b_clust_best for x in labels]
    # Update labels (cluster labels are given in descending order of cluster size)
    sorted_labels = pd.Series(labels).value_counts().index  # sorted in descending order of size
    dict_update = {label: i for label, i in zip(sorted_labels, range(0, len(set(labels))))}
    labels = [dict_update[label] for label in labels]
    return labels




"""
This is a script for the frontend of the CPP class, a sequence-based feature engineering object.
"""
import numpy as np
import pandas as pd
from typing import Optional, List, Union

import aaanalysis.utils as ut
from aaanalysis.template_classes import Tool

# Import supportive class (exception for importing from same sub-package)
from ._backend.cpp.sequence_feature import get_features_, get_split_kws_
from ._backend.check_feature import (check_split_kws,
                                     check_parts_len,
                                     check_match_features_seq_parts,
                                     check_df_parts,
                                     check_match_df_parts_features,
                                     check_match_df_parts_list_parts,
                                     check_match_df_parts_split_kws,
                                     check_df_scales,
                                     check_match_df_scales_features,
                                     check_df_cat,
                                     check_match_df_cat_features,
                                     check_match_df_parts_df_scales,
                                     check_match_df_scales_df_cat)
from ._backend.cpp.utils_feature import get_positions_, add_scale_info_
from ._backend.cpp.cpp_run import pre_filtering_info, pre_filtering, filtering, add_stat
from ._backend.cpp.cpp_eval import evaluate_features


# I Helper Functions
def check_sample_in_df_seq(sample_name=None, df_seq=None):
    """Check if sample name in df_seq"""
    list_names = list(df_seq[ut.COL_NAME])
    if sample_name not in list_names:
        error = f"'sample_name' ('{sample_name}') not in '{ut.COL_NAME}' of 'df_seq'." \
                f"\nValid names are: {list_names}"
        raise ValueError(error)


def check_match_list_df_feat_list_df_parts(list_df_feat=None, list_df_parts=None):
    """Check if all elements in list are valid feature DataFrames"""
    for df_feat, df_parts in zip(list_df_feat, list_df_parts):
        ut.check_df_feat(df_feat=df_feat, list_parts=list(df_parts))


def check_match_list_df_feat_names_feature_sets(list_df_feat=None, names_feature_sets=None):
    """Check if length of list_df_feat and names match"""
    if names_feature_sets is None:
        return None # Skip check
    if len(list_df_feat) != len(names_feature_sets):
        raise ValueError(f"Length of 'list_df_feat' ({len(list_df_feat)}) and 'names_feature_sets'"
                         f" ({len(names_feature_sets)} does not match) ")


# II Main Functions
class CPP(Tool):
    """
    Comparative Physicochemical Profiling (**CPP**) class to create and filter features that are most discriminant
    between two sets of sequences [Breimann24c]_.

    CPP aims at identifying a set of non-redundant features that are most discriminant between the
    test and reference group of sequences.

    Attributes
    ----------
    df_parts
        DataFrame with sequence **Parts**.
    split_kws
        Nested dictionary defining **Splits** with parameter dictionary for each chosen split_type.
    df_scales
        DataFrame with amino acid **Scales**.
    df_cat
        DataFrame with categories for physicochemical amino acid **Scales**.
    """
    def __init__(self,
                 df_parts: pd.DataFrame = None,
                 split_kws: Optional[dict] = None,
                 df_scales: Optional[pd.DataFrame] = None,
                 df_cat: Optional[pd.DataFrame] = None,
                 accept_gaps: bool = False,
                 verbose: bool = True,
                 random_state: Optional[str] = None,
                 ):
        """
        Parameters
        ----------
        df_parts : pd.DataFrame, shape (n_samples, n_parts)
            DataFrame with sequence parts.
        split_kws : dict, optional
            Dictionary with parameter dictionary for each chosen split_type. Default from :meth:`SequenceFeature.get_split_kws`.
        df_scales : pd.DataFrame, shape (n_letters, n_scales), optional
            DataFrame of scales with letters typically representing amino acids. Default from :meth:`load_scales`
            unless specified in ``options['df_scales']``.
        df_cat : pd.DataFrame, shape (n_scales, n_scales_info), optional
            DataFrame of categories for physicochemical scales. Must contain all scales from ``df_scales``.
            Default from :meth:`load_scales` with ``name='scales_cat'``, unless specified in ``options['df_cat']``.
        accept_gaps : bool, default=False
            Whether to accept missing values by enabling omitting for computations (if ``True``).
        verbose : bool, default=True
            If ``True``, verbose outputs are enabled.
        random_state : int, optional
            The seed used by the random number generator. If a positive integer, results of stochastic processes are
            consistent, enabling reproducibility. If ``None``, stochastic processes will be truly random.

        Notes
        -----
        * All scales from ``df_scales`` must be contained in ``df_cat``

        See Also
        --------
        * :class:`CPPPlot`: the respective plotting class.
        * :class:`SequenceFeature` for definition of sequence ``Parts``.
        * :meth:`SequenceFeature.split_kws` for definition of ``Splits`` key word arguments.
        * :func:`load_scales` for definition of amino acid ``Scales`` and their categories.

        Examples
        --------
        .. include:: examples/cpp.rst
        """
        # Load defaults
        if split_kws is None:
            split_kws = get_split_kws_()
        if df_scales is None:
            df_scales = ut.load_default_scales()
        if df_cat is None:
            df_cat = ut.load_default_scales(scale_cat=True)
        # Check input
        verbose = ut.check_verbose(verbose)
        random_state = ut.check_random_state(random_state=random_state)
        check_df_parts(df_parts=df_parts)
        check_split_kws(split_kws=split_kws)
        check_df_scales(df_scales=df_scales)
        check_df_cat(df_cat=df_cat)
        ut.check_bool(name="accept_gaps", val=accept_gaps)
        df_parts = check_match_df_parts_df_scales(df_parts=df_parts, df_scales=df_scales, accept_gaps=accept_gaps)
        check_match_df_parts_split_kws(df_parts=df_parts, split_kws=split_kws)
        df_scales, df_cat = check_match_df_scales_df_cat(df_cat=df_cat, df_scales=df_scales, verbose=verbose)
        # Internal attributes
        self._accept_gaps = accept_gaps
        self._verbose = verbose
        self._random_state = random_state
        # Feature components: Scales + Part + Split
        self.df_cat = df_cat.copy()
        self.df_scales = df_scales.copy()
        self.df_parts = df_parts.copy()
        self.split_kws = split_kws

    # Main method
    def run(self,
            labels: ut.ArrayLike1D = None,
            label_test: int = 1,
            label_ref: int = 0,
            n_filter: int = 100,
            n_pre_filter: Optional[int] = None,
            pct_pre_filter: int = 5,
            max_std_test: float = 0.2,
            max_overlap: float = 0.5,
            max_cor: float = 0.5,
            check_cat: bool = True,
            parametric: bool = False,
            start: int = 1,
            tmd_len: int = 20,
            jmd_n_len: int = 10,
            jmd_c_len: int = 10,
            n_jobs: Optional[int] = None
            ) -> pd.DataFrame:
        """
        Perform Comparative Physicochemical Profiling (CPP) algorithm: creation and two-step filtering of
        interpretable sequence-based features.

        The aim of the CPP algorithm is to identify a set of unique, non-redundant features that are most
        discriminant between the test and reference group of sequences. See [Breimann24c]_ for details on the algorithm.

        Parameters
        ----------
        labels : array-like, shape (n_samples,)
            Class labels for samples in sequence DataFrame (typically, test=1, reference=0).
        label_test : int, default=1,
            Class label of test group in ``labels``.
        label_ref : int, default=0,
            Class label of reference group in ``labels``.
        n_filter : int, default=100
            Number of features to be filtered/selected by CPP algorithm.
        n_pre_filter : int, optional
            Number of feature to be pre-filtered by CPP algorithm. If ``None``, a percentage of all features is used.
        pct_pre_filter : int, default=5
            Percentage of all features that should remain after the pre-filtering step.
        max_std_test : float, default=0.2
            Maximum standard deviation [>0-<1] within the test group used as threshold for pre-filtering.
        max_overlap : float, default=0.5
            Maximum positional overlap [0-1] of features used as threshold for filtering.
        max_cor : float, default=0.5
            Maximum Pearson correlation [0-1] of feature scales used as threshold for filtering.
        check_cat : bool, default=True
            Whether to check for redundancy within scale categories during filtering.
        parametric : bool, default=False
            Whether to use parametric (T-test) or non-parametric (Mann-Whitney-U-test) test for p-value computation.
        start : int, default=1
            Position label of first residue position (starting at N-terminus).
        tmd_len : int, default=20
            Length of TMD (>0).
        jmd_n_len : int, default=10
            Length of JMD-N (>=0).
        jmd_c_len : int, default=10
            Length of JMD-C (>=0).
        n_jobs : int, None, or -1, default=None
            Number of CPU cores used for multiprocessing. If ``None``, the number is optimized automatically.
            If ``-1``, the number is set to all available cores.

        Returns
        -------
        df_feat : pd.DataFrame, shape (n_features, n_feature_info)
            Feature DataFrame with a unique identifier, scale information, statistics, and positions for each feature.

        Notes
        -----
        * Pre-filtering can be adjusted by the following parameters: {'n_pre_filter', 'pct_pre_filter', 'max_std_test'}.
        * Filtering can be adjusted by the following parameters: {'n_filter', 'max_overlap', 'max_cor', 'check_cat'}.
        * ``check_cat`` (not introduced in [Breimann24c]_) was included to provide higher filtering flexibility.
        * ``df_feat`` contains the following 13 columns, including the unique feature id (1), scale information (2-5),
           statistical results for filtering and ranking (6-12), and feature positions (13):

            1. 'features': Feature ID (PART-SPLIT-SCALE)
            2. 'category': Scale category
            3. 'subcategory': Sub category of scales
            4. 'scale_name': Name of scales
            5. 'scale_description': Description of the scale
            6. 'abs_auc': Absolute adjusted AUC (area under the curve) [-0.5 to 0.5]
            7. 'abs_mean_dif': Absolute mean differences between test and reference group [0 to 1]
            8. 'mean_dif': Mean differences between test and reference group [-1 to 1]
            9. 'std_test': Standard deviation in test group
            10. 'std_ref': Standard deviation in reference group
            11. 'p_val': Non-parametric (mann_whitney) or parametric (ttest_indep) statistic
            12. 'p_val_fdr_bh': Benjamini-Hochberg FDR corrected p-values
            13. 'positions': Feature positions for default settings

        See Also
        --------
        * :func:`comp_auc_adjusted` for details on 'abs_auc'.

        Examples
        --------
        .. include:: examples/cpp_run.rst
        """
        # Check input
        ut.check_number_val(name="label_test", val=label_test, just_int=True)
        ut.check_number_val(name="label_ref", val=label_ref, just_int=True)
        labels = ut.check_labels(labels=labels, vals_requiered=[label_test, label_ref],
                                 len_requiered=len(self.df_parts), allow_other_vals=False)
        ut.check_number_range(name="n_filter", val=n_filter, min_val=1, just_int=True)
        ut.check_number_range(name="n_pre_filter", val=n_pre_filter, min_val=1, accept_none=True, just_int=True)
        ut.check_number_range(name="pct_pre_filter", val=pct_pre_filter, min_val=5, max_val=100, just_int=True)
        ut.check_number_range(name="max_std_test", val=max_std_test, min_val=0.0, max_val=1.0,
                              just_int=False, exclusive_limits=True)
        ut.check_number_range(name="max_overlap", val=max_overlap, min_val=0.0, max_val=1.0, just_int=False)
        ut.check_number_range(name="max_cor", val=max_cor, min_val=0.0, max_val=1.0, just_int=False)
        ut.check_bool(name="check_cat", val=check_cat)
        ut.check_bool(name="parametric", val=parametric)
        ut.check_number_val(name="start", val=start, just_int=True, accept_none=False)
        args_len, _ = check_parts_len(tmd_len=tmd_len, jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len)
        n_jobs = ut.check_n_jobs(n_jobs=n_jobs)
        # Settings and creation of objects
        n_feat = len(get_features_(list_parts=list(self.df_parts),
                                   split_kws=self.split_kws,
                                   list_scales=list(self.df_scales)))
        n_filter = n_feat if n_feat < n_filter else n_filter
        if self._verbose:
            start_message = f"1. CPP creates {n_feat} features for {len(self.df_parts)} samples"
            ut.print_start_progress(start_message=start_message)
        # Pre-filtering: Select best n % of feature (filter_pct) based std(test set) and mean_dif
        abs_mean_dif, std_test, features = pre_filtering_info(df_parts=self.df_parts,
                                                              split_kws=self.split_kws,
                                                              df_scales=self.df_scales,
                                                              labels=labels,
                                                              label_test=label_test,
                                                              label_ref=label_ref,
                                                              accept_gaps=self._accept_gaps,
                                                              verbose=self._verbose,
                                                              n_jobs=n_jobs)
        n_feat = int(len(features))
        if n_pre_filter is None:
            n_pre_filter = int(n_feat * (pct_pre_filter / 100))
            n_pre_filter = n_filter if n_pre_filter < n_filter else n_pre_filter
        if self._verbose:
            pct_pre_filter = np.round((n_pre_filter/n_feat*100), 2)
            end_message = (f"2. CPP pre-filters {n_pre_filter} features ({pct_pre_filter}%) with highest "
                           f"'{ut.COL_ABS_MEAN_DIF}' and 'max_std_test' <= {max_std_test}")
            ut.print_end_progress(end_message=end_message)
        df = pre_filtering(features=features,
                           abs_mean_dif=abs_mean_dif,
                           std_test=std_test,
                           n=n_pre_filter,
                           max_std_test=max_std_test)
        features = df[ut.COL_FEATURE].to_list()
        # Add feature information
        df = add_stat(df_feat=df, df_scales=self.df_scales, df_parts=self.df_parts,
                      labels=labels, parametric=parametric, accept_gaps=self._accept_gaps,
                      label_test=label_test, label_ref=label_ref, n_jobs=n_jobs)
        feat_positions = get_positions_(features=features, start=start, **args_len)
        df[ut.COL_POSITION] = feat_positions
        df = add_scale_info_(df_feat=df, df_cat=self.df_cat)
        # Filtering using CPP algorithm
        if self._verbose:
            ut.print_out(f"3. CPP filtering algorithm")
        df_feat = filtering(df=df, df_scales=self.df_scales,
                            n_filter=n_filter, check_cat=check_cat,
                            max_overlap=max_overlap, max_cor=max_cor)
        # Adjust df_feat
        df_feat.reset_index(drop=True, inplace=True)
        df_feat[ut.COLS_FEAT_STAT] = df_feat[ut.COLS_FEAT_STAT].round(3)
        if self._verbose:
            ut.print_out(f"4. CPP returns df with {len(df_feat)} unique features including general information and statistics")
        return df_feat

    def eval(self,
             list_df_feat: List[pd.DataFrame] = None,
             labels: ut.ArrayLike1D = None,
             label_test: int = 1,
             label_ref: int = 0,
             min_th: float = 0.0,
             names_feature_sets: Optional[List[str]] = None,
             list_cat: Optional[List[str]] = None,
             list_df_parts: Optional[List[pd.DataFrame]] = None,
             n_jobs: Optional[int] = 1,
             ) -> pd.DataFrame:
        """
        Evaluate the quality of different sets of identified CPP features.

        Feature sets are evaluated regarding two quality groups:

        - **Discriminative Power**: The capability of features to distinguish between test and reference datasets.
        - **Redundancy**: Assessed by the optimized number of clusters, based on Pearson correlation among features.

        Parameters
        ----------
        list_df_feat : list of pd.DataFrames
            List of feature DataFrames each of shape (n_features, n_feature_info)
        labels : array-like, shape (n_samples,)
            Class labels for samples in sequence DataFrame (typically, test=1, reference=0).
        label_test : int, default=1,
            Class label of test group in ``labels``.
        label_ref : int, default=0,
            Class label of reference group in ``labels``.
        min_th : float, default=0.0
            Pearson correlation threshold for clustering optimization (between -1 and 1).
        names_feature_sets : list of str, optional
            List of names for feature sets corresponding to ``list_df_feat``.
        list_cat : list of str, optional
            List of scale categories to retrieve number of features from. Default:
            ['ASA/Volume', 'Composition', 'Conformation', 'Energy', 'Others', 'Polarity', 'Shape', 'Structure-Activity']
        list_df_parts : list of pd.DataFrames, optional
            List of part DataFrames each of shape (n_samples, n_parts). Must match with ``list_df_feat``.
        n_jobs : int, None, or -1, default=1
            Number of CPU cores used for multiprocessing. If ``None``, the number is optimized automatically.
            If ``-1``, the number is set to all available cores.

        Returns
        -------
        df_eval : pd.DataFrame
            Evaluation results for each set of identified features. For each set, statistical
            measures were averaged across all features.

        Notes
        -----
        * ``df_eval`` includes the following columns (upper-case indicates direct reference to ``df_feat`` columns):

            - 'name': Name of the feature set, typically based on CPP run settings, if ``names`` is provided.
            - 'n_features': Tuple with total number of features and list of number of features per scale category from ``list_cat``.
            - 'avg_ABS_AUC': Absolute AUC averaged across all features.
            - 'range_ABS_AUC': Quintile range of absolute AUC among all features (min, 25%, median, 75%, max).
            - 'avg_MEAN_DIF': Tuple of mean differences averaged across all features separately
              for features with positive and negative 'mean_dif'.
            - 'n_clusters': Optimal number of clusters [2,100].
            - 'avg_n_feat_per_clust': Average number of features per cluster.
            - 'std_n_feat_per_clust': Standard deviation of feature number per cluster.

        * 'n_clusters' is optimized for a KMeans clustering model based on the minimum Pearson correlation between
          the cluster center and all cluster members across all clusters (``min_cor_center`` in :class:`AAclust`),
          which has to exceed the minimum correlation threshold ``min_th``.

        See Also
        --------
        * :func:`CPPPlot.eval`: the respective plotting method.
        * :ref:`usage_principles_aaontology` for details on scale categories.
        * :meth:`CPP.run` for details on CPP statistical measures.
        * :func:`comp_auc_adjusted` for details on 'abs_auc'.
        * :class:`sklearn.cluster.KMeans` for employed clustering model.
        * :class:`AAclust` ([Breimann24a]_) for details on cluster optimization using Pearson correlation.

        Examples
        --------
        .. include:: examples/cpp_eval.rst
        """
        # Check input
        list_df_feat = ut.check_list_like(name="list_df_feat", val=list_df_feat, min_len=2)
        ut.check_number_val(name="label_test", val=label_test, just_int=True)
        ut.check_number_val(name="label_ref", val=label_ref, just_int=True)
        labels = ut.check_labels(labels=labels, vals_requiered=[label_test, label_ref],
                                 len_requiered=len(self.df_parts), allow_other_vals=False)
        ut.check_number_range(name="min_th", val=min_th, min_val=-1, max_val=1, just_int=False)
        names_feature_sets = ut.check_list_like(name="names_feature_sets", val=names_feature_sets, accept_none=True,
                                            accept_str=True, check_all_str_or_convertible=True)
        list_cat = ut.check_list_like(name="list_cat", val=list_cat, accept_none=True, accept_str=True,
                                      check_all_str_or_convertible=True)
        n_jobs = ut.check_n_jobs(n_jobs=n_jobs)
        check_match_list_df_feat_names_feature_sets(list_df_feat=list_df_feat,
                                                    names_feature_sets=names_feature_sets)
        ut.check_list_like(name="list_df_parts", val=list_df_parts, accept_none=True)
        mask_test = [x == label_test for x in labels]
        if list_df_parts is None:
            list_df_parts = [self.df_parts[mask_test]] * len(list_df_feat)
        if list_cat is None:
            list_cat = ut.LIST_CAT
        check_match_list_df_feat_list_df_parts(list_df_feat=list_df_feat, list_df_parts=list_df_parts)
        df_eval = evaluate_features(list_df_feat=list_df_feat,
                                    names_feature_sets=names_feature_sets,
                                    list_cat=list_cat,
                                    list_df_parts=list_df_parts,
                                    df_scales=self.df_scales,
                                    accept_gaps=self._accept_gaps,
                                    min_th=min_th,
                                    n_jobs=n_jobs,
                                    random_state=self._random_state)
        return df_eval



"""
This is a script for creating template Wrapper and Tool classes used in the AAanalysis library
"""
from abc import ABC, abstractmethod


class Tool(ABC):
    """Tool class for specialized tasks.

    This class provides a framework for standalone utilities focusing on specialized tasks,
    such as feature engineering or data pre-processing. The `.run` method is intended for
    executing the main logic of the tool, while `.eval` is for evaluation of the tool's output.

    Subclasses should provide concrete implementations for the `run` and `eval` methods.
    """

    @abstractmethod
    def run(self):
        """Execute the main logic of the tool.

        This method should be overridden by subclasses to implement the main functionality.

        Raises:
        NotImplementedError: If the method is not implemented by the subclass.
        """
        raise NotImplementedError

    @abstractmethod
    def eval(self):
        """Evaluate the output generated by the tool."""
        raise NotImplementedError


class Wrapper(ABC):
    """Wrapper class for extending models from libraries like scikit-learn."""

    @abstractmethod
    def __init__(self):
        # Model attributes to be implemented by wrapper class
        self._model_class = None
        self.model = None
        self.model_kwargs = None

    @abstractmethod
    def fit(self, *args, **kwargs):
        """Fit the model with data."""

    @abstractmethod
    def eval(self, *args, **kwargs):
        """Evaluate the model."""
        raise NotImplementedError



"""
This is a script for scale loading function. Please define new loading functions by their loaded data by introducing
 a new data table in docs/source/index/tables_templates.rst.
"""
from typing import Optional, Union
from pandas import DataFrame
from aaanalysis import utils as ut


# Check functions for load_scales
def check_name_of_scale(name=None):
    # Check if the provided scale name is valid
    if name not in ut.NAMES_SCALE_SETS:
        raise ValueError(f"'name' ('{name}') is not valid. Choose one of following: {ut.NAMES_SCALE_SETS}")


def check_top60_n(name=None, top60_n=None):
    """Check if name is valid and top60_n is between 1 and 60"""
    if top60_n is None:
        return
    if isinstance(top60_n, str):
        if "AAC" not in top60_n:
            raise ValueError(f"'top60_n' ('{top60_n}') should be int or 'AAC' id")
        top60_n = int(top60_n.replace("AAC", ""))
    ut.check_number_range(name="top60_n", val=top60_n, min_val=1, max_val=60, just_int=True)
    matching_scale_sets = [ut.STR_SCALES, ut.STR_SCALE_CAT, ut.STR_SCALES_RAW]
    if name not in matching_scale_sets:
        raise ValueError(f"'name' ('{name}') is not valid for 'top60_n' ({top60_n})."
                         f" Choose one of following: {matching_scale_sets}")
    return top60_n


# Helper functions for load_scales
def _filter_scales(df_cat=None, unclassified_out=False, just_aaindex=False):
    """Filter scales for unclassified and aaindex scales"""
    list_ids_not_in_aaindex = [x for x in df_cat[ut.COL_SCALE_ID] if "LINS" in x or "KOEH" in x]
    list_ids_unclassified = [x for x, cat, sub_cat in zip(df_cat[ut.COL_SCALE_ID], df_cat[ut.COL_CAT], df_cat[ut.COL_SUBCAT])
                             if "Unclassified" in sub_cat or cat == "Others"]
    list_ids_to_exclude = []
    if unclassified_out:
        list_ids_to_exclude.extend(list_ids_unclassified)
    if just_aaindex:
        list_ids_to_exclude.extend(list_ids_not_in_aaindex)
    df_cat = df_cat[~df_cat[ut.COL_SCALE_ID].isin(list_ids_to_exclude)]
    return df_cat


def _get_selected_scales(top60_n=None):
    """Get selected top scale set"""
    df_eval = ut.read_excel_cached(ut.FOLDER_DATA + f"{ut.STR_TOP60}.xlsx").drop("top60_id", axis=1)
    # Find the names of the index where the value is not 0
    _df = df_eval.iloc[top60_n - 1]
    selected_scales = _df[_df != 0].index.tolist()
    return selected_scales


def _adjust_dtypes(df=None, name=None):
    """Set dtypes to avoid loading problems"""
    # Adjust data type of column values
    name_all_float = [ut.STR_SCALES, ut.STR_SCALES_RAW, ut.STR_SCALES_PC]
    name_all_int = [ut.STR_TOP60]
    if name in name_all_float:
        df = df.astype(float)
    elif name in name_all_int:
        df = df.astype(int)
    return df


# II Main Functions
def load_scales(name: str = "scales",
                just_aaindex: bool = False,
                unclassified_out: bool = False,
                top60_n: Optional[Union[int, str]] = None
                ) -> DataFrame:
    """
    Load amino acid scales or their classification (AAontology).

    The amino acid scales (``name='scales_raw'``) encompass all scales from AAindex ([Kawashima08]_) along with two
    additional data sources. These scales were min-max normalized (``name='scales'``) and organized in a two-level
    classification called AAontology (``name='scales_cat'``), as detailed in [Breimann24b]_. The first 20 principal
    components (PCs) of all compressed scales are provided (``name='scales_pc'``) and were used for an in-depth analysis
    of redundancy-reduced scale subsets obtained by :class:`AAclust` ([Breimann24a]_). The top 60 scale sets from
    this analysis are available either collectively (``name='top60'``) or individually (``top60_n='1-60'``),
    accompanied by their evaluations in the ``'top60_eval'`` dataset.

    Parameters
    ----------
    name : str, default='scales'
        Name of the loaded dataset:

        - ``scales_raw``: All amino acid scales.
        - ``scales``:  Min-max normalized raw scales.
        - ``scales_cat``: Two-level classification (AAontology).
        - ``scales_pc``: First 20 PCs of compressed scales.
        - ``top60``:  Selection of 60 best performing scale sets.
        - ``top60_eval``: Evaluation of 60 best performing scale sets.

        Or Number between 1 and 60 to select the i-th top60 dataset.

    just_aaindex : bool, default=False
        If ``True``, returns only scales from AAindex. Relevant only for 'scales', 'scales_raw', or 'scales_cat'.
    unclassified_out : bool, default=False
        Determines exclusion of unclassified scales. Relevant only for 'scales', 'scales_raw', or 'scales_cat'.
    top60_n : int or str, optional
         Select the n-th scale set from top60 sets and return it for 'scales', 'scales_raw', or 'scales_cat'.
         Allowed strings are AAclust ids (e.g., 'AAC01').

    Returns
    -------
    pandas.DataFrame
        DataFrame containing the chosen dataset, recommended to be named by their name suffix (``df_scales``, ``df_cat``).

    Notes
    -----
    * ``df_cat`` includes the following columns:

        - 'scale_id': ID of scale (from AAindex or following the same naming convention).
        - 'category': Category of scale (defined in AAontology).
        - 'subcategory': Subcategory of scale (AAontology).
        - 'scale_name': Name of scale derived from scale description.
        - 'scale_description': Description of scale (derived from AAindex).

    * Scales under the 'Others' category are considered unclassified.

    See Also
    --------
    * Overview of all loading options: :ref:`t2_overview_scales`.
    * AAontology: :ref:`t3a_aaontology_categories` and :ref:`t3b_aaontology_subcategories` tables.
    * Step-by-step guide in the `Scale Loading Tutorial <tutorial2b_scales_loader.html>`_.
    * :class:`AAclust` for customizing redundancy-reduced scale sets.

    Examples
    --------
    .. include:: examples/load_scales.rst
    """
    # DEV: every returned must be copied to avoid in-place mutation for newly loaded dataframes
    check_name_of_scale(name=name)
    ut.check_bool(name="just_aaindex", val=just_aaindex)
    ut.check_bool(name="unclassified_in", val=unclassified_out)
    top60_n = check_top60_n(name=name, top60_n=top60_n)
    # Load and filter top60 scales
    if top60_n is not None:
        selected_scales = _get_selected_scales(top60_n=top60_n)
        if name == ut.STR_SCALE_CAT:
            df_cat = ut.read_excel_cached(ut.FOLDER_DATA + f"{ut.STR_SCALE_CAT}.xlsx")
            df_cat = df_cat[df_cat[ut.COL_SCALE_ID].isin(selected_scales)].reset_index(drop=True)
            return df_cat.copy()
        elif name in [ut.STR_SCALES, ut.STR_SCALES_RAW]:
            df = ut.read_excel_cached(ut.FOLDER_DATA + name + ".xlsx", index_col=0)
            df = df[selected_scales]
            df = _adjust_dtypes(df=df, name=name)
            return df.copy()
        else:
            raise ValueError(f"Wrong 'name' ('{name}') for 'top60_n")

    # Load unfiltered data
    if not unclassified_out and not just_aaindex:
        if name == ut.STR_SCALE_CAT:
            df_cat = ut.read_excel_cached(ut.FOLDER_DATA + f"{ut.STR_SCALE_CAT}.xlsx")
            return df_cat
        df = ut.read_excel_cached(ut.FOLDER_DATA + name + ".xlsx", index_col=0)
        df = _adjust_dtypes(df=df, name=name)
        return df.copy()

    # Load and filter scale categories
    df_cat = ut.read_excel_cached(ut.FOLDER_DATA + f"{ut.STR_SCALE_CAT}.xlsx")
    df_cat = _filter_scales(df_cat=df_cat, unclassified_out=unclassified_out, just_aaindex=just_aaindex)
    if name == ut.STR_SCALE_CAT:
        return df_cat.reset_index(drop=True).copy()

    # Load and filter scales
    df = ut.read_excel_cached(ut.FOLDER_DATA + name + ".xlsx", index_col=0)
    if name in [ut.STR_SCALES, ut.STR_SCALES_RAW]:
        selected_scales = [x for x in list(df) if x in list(df_cat[ut.COL_SCALE_ID])]
        df = df[selected_scales]
    df = _adjust_dtypes(df=df, name=name)
    return df.copy()



"""
This is a script for reading and writing df_seq to the fasta file format, which is the most commonly used format
in computational biology. This fasta format enables a smooth interaction with the biopython package.
See https://bioperl.org/formats/sequence_formats/FASTA_sequence_format for description of input fasta format.
"""
import time
import pandas as pd
import numpy as np

import aaanalysis.utils as ut



# I Helper Functions
# TODO add more parsers for often used data formats in computational biology (make overview)
# TODO implement parser from df to df_seq (remove not necessary columns and adjust naming)


# II Main Functions
def read_fasta():
    """"""


def to_fasta(df=None, fasta_name=None, col_id=None, col_seq=None, cols_info=None):
        """"""
        # Check input
        ut.check_str(name="col_id", val=col_id, accept_none=False)
        ut.check_str(name="col_seq", val=col_seq, accept_none=False)
        cols_info = ut.check_list_like(name="cols_info", val=cols_info, accept_str=True, accept_none=True)
        cols_requiered = [col_id, col_seq]
        if cols_info is not None:
            cols_requiered += cols_info
        ut.check_df(df=df, name="df", cols_requiered=cols_requiered, accept_none=False, accept_nan=False)
        # Create faste
        if ".fasta" not in fasta_name:
            fasta_name += ".fasta"
        fasta = open(fasta_name, "w")
        for i, row in df.iterrows():
            seq = row[col_seq]
            entry = row[col_id]
            if cols_info is not None:
                if isinstance(cols_info, list):
                    str_info = ", ".join([str(row[c]) for c in cols_info])
                else:
                    str_info = row[cols_info]
                fasta.write(">" + entry + "," + str_info + "\n" + seq + "\n")
            else:
                fasta.write(">" + entry + "\n" + seq + "\n")
        fasta.close()



def to_df_seq(df=None):
    """"""



from ._load_dataset import load_dataset
from ._load_scales import load_scales
from ._load_features import load_features
from ._data_read_write import to_fasta

__all__ = [
    "load_dataset",
    "load_scales",
    "load_features",
    "to_fasta",
]



"""
This is a script for the frontend of the AAclustPlot class, used for plotting of the AAclust results.
"""
import pandas as pd
from sklearn.decomposition import PCA
from typing import Optional, Dict, Union, List, Tuple, Type
from sklearn.base import TransformerMixin
import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np
import warnings

import aaanalysis.utils as ut

from ._backend.check_aaclust import check_metric
from ._backend.aaclust.aaclust_plot import plot_eval, plot_center_or_medoid, plot_correlation


# I Helper Functions
def check_match_df_eval_names(df_eval=None, names=None):
    """Validate the match between the number of samples in df_eval and the length of names"""
    n_samples = len(df_eval)
    if names is not None:
        if len(names) != n_samples:
            raise ValueError(f"n_samples does not match for 'data' ({n_samples}) and 'names' ({len(names)}).")
    else:
        names = [f"Set {i}" for i in range(1, n_samples + 1)]
    df_eval.index = names
    return df_eval


# Check correlation plot
def check_match_df_corr_labels(df_corr=None, labels=None):
    """Ensure the number of labels matches the number of samples in df_corr"""
    labels = ut.check_labels(labels=labels)
    n_samples, n_clusters = df_corr.shape
    if n_samples != len(labels):
        raise ValueError(f"Number of 'labels' ({len(labels)}) must match with n_samples in 'df_corr' ({n_samples})")
    return labels

def check_match_df_corr_labels_ref(df_corr=None, labels_ref=None, labels=None, pairwise=False, verbose=False):
    """Ensure the number of labels matches the number of samples in df_corr"""
    str_add = ""
    if labels_ref is None:
        if pairwise:
            return # Skip check
        columns = list(df_corr)
        if sum([isinstance(i, str) for i in columns]) != 0:
            raise ValueError("'labels_ref' must be provided if columns in 'df_corr' and strings.")
        labels_ref = list(df_corr)
        str_add = "(obtained from 'df_corr' columns if not given)"
    if set(labels) != set(labels_ref) and verbose:
        warnings.warn(f"'labels' and 'labels_ref' {str_add} does not match. Provide 'labels_ref' or adjust 'df_corr' columns."
                      f"\n 'labels': {set(labels)},"
                      f"\n 'labels_ref': {set(labels_ref)}")

    labels_ref = ut.check_labels(labels=labels_ref, name="labels_ref")
    n_samples, n_clusters = df_corr.shape
    if n_clusters != len(labels_ref):
        raise ValueError(f"Number of 'labels_ref' ({len(labels_ref)}) must match with n_clusters in 'df_corr' ({n_samples})")
    return labels_ref


def check_method(method=None):
    """Validate the method parameter against a list of valid hierarchical clustering methods"""
    valid_methods = ["single", "complete", "average", "weighted", "centroid", "median", "ward"]
    if method not in valid_methods:
        raise ValueError(f"'method' ({method}) should be one of following: {valid_methods}")


def check_match_df_corr_clust_x(df_corr=None, cluster_x=None):
    """Ensure df_corr has variability and no missing values if cluster_x is enabled"""
    all_vals = df_corr.to_numpy().flatten()[1:].tolist()
    if cluster_x:
        if len(set(all_vals)) == 1:
            raise ValueError(f"'df_corr' should not contain all same values if 'cluster_x' is True")
        if None in all_vals or np.NaN in all_vals:
            raise ValueError(f"'df_corr' should not contain missing values")


def check_bar_position(bar_position=None):
    """Validate and standardize bar_position to be one of the valid positions like 'left', 'right', 'top', or 'bottom'"""
    bar_position = ut.check_list_like(name="bar_position", val=bar_position, convert=True, accept_str=True)
    valid_positions = ['left', 'right', 'top', 'bottom']
    wrong_positions = [x for x in bar_position if x not in valid_positions]
    if len(wrong_positions) > 0:
        raise ValueError(f"Wrong 'bar_position' ({wrong_positions}). They should be as follows: {valid_positions}")
    return bar_position


def check_bar_colors(bar_colors=None):
    """Validate and standardize bar_colors to be either a single color or a list of valid colors."""
    if isinstance(bar_colors, str):
        ut.check_color(name='bar_colors', val=bar_colors, accept_none=False)
        bar_colors = [bar_colors]
    elif isinstance(bar_colors, list):
        for color in bar_colors:
            ut.check_color(name="element from 'bar_colors'", val=color, accept_none=False)
    else:
        raise ValueError("'bar_colors' should be string or list")
    return bar_colors


def check_match_bar_colors_labels(bar_colors=None, labels=None):
    """Adjust bar_colors to match the number of unique labels and warn if there are insufficient colors provided"""
    n_clusters = len(set(labels)) # Number of unique labels
    if len(bar_colors) == 1:
        bar_colors = bar_colors * n_clusters
    if len(bar_colors) < n_clusters:
        n_colors = len(bar_colors)
        bar_colors *= n_clusters
        warnings.warn(f"Length of 'bar_colors' (n={n_colors}) should be >= n_clusters (n={n_clusters})")
    return bar_colors[0:n_clusters]


# II Main Functions
class AAclustPlot:
    """
    Plotting class for :class:`AAclust` (Amino Acid clustering) results [Breimann24a]_.

    This class performs dimensionality reduction for visualization using decomposition models such as
    Principal Component Analysis (PCA).
    """
    def __init__(self,
                 model_class: Type[TransformerMixin] = PCA,
                 model_kwargs: Optional[Dict] = None,
                 verbose: bool = True,
                 random_state: Optional[str] = None,
                 ):
        """
        Parameters
        ----------
        model_class : Type[TransformerMixin], default=PCA
            A decomposition model class with ``n_components`` parameter.
        model_kwargs : dict, optional
            Keyword arguments to pass to the selected decomposition model.
        verbose : bool, default=True
            If ``True``, verbose outputs are enabled.
        random_state : int, optional
            The seed used by the random number generator. If a positive integer, results of stochastic processes are
            consistent, enabling reproducibility. If ``None``, stochastic processes will be truly random.

        See Also
        --------
        * :class:`AAclust`: the respective computation class.
        * Scikit-learn `decomposition model classes <https://scikit-learn.org/stable/modules/classes.html#module-sklearn.decomposition>`_.

        Examples
        --------
        .. include:: examples/aac_plot.rst
        """
        # Global parameters
        verbose = ut.check_verbose(verbose)
        random_state = ut.check_random_state(random_state=random_state)
        # Model parameters
        ut.check_mode_class(model_class=model_class)
        model_kwargs = ut.check_model_kwargs(model_class=model_class,
                                             model_kwargs=model_kwargs,
                                             param_to_check="n_components",
                                             method_to_check="transform",
                                             random_state=random_state)
        # Internal attributes
        self._verbose = verbose
        self._random_state = random_state
        self._model_class = model_class
        self._model_kwargs = model_kwargs

    @staticmethod
    def eval(df_eval: pd.DataFrame = None,
             figsize: Tuple[Union[int, float], Union[int, float]] = (6, 4),
             dict_xlims: Optional[Union[None, dict]] = None,
             ) -> Tuple[plt.Figure, plt.Axes]:
        """
        Plots evaluation of ``n_clusters`` and clustering metrics ``BIC``, ``CH``, and ``SC`` from ``df_seq``.

        The clustering evaluation metrics (BIC, CH, and SC) are ranked by the average of their independent rankings.

        Parameters
        ----------
        df_eval : pd.DataFrame, shape (n_datasets, n_metrics)
            DataFrame with evaluation measures for scale sets. Each `row` corresponds to a specific scale set
            and `columns` are as follows:

            - 'name': Name of clustering datasets.
            - 'n_clusters': Number of clusters.
            - 'BIC': Bayesian Information Criterion.
            - 'CH': Calinski-Harabasz Index.
            - 'SC': Silhouette Coefficient.

        figsize : tuple, default=(7, 6)
            Figure dimensions (width, height) in inches.
        dict_xlims : dict, optional
            A dictionary containing x-axis limits for subplots. Keys should be the subplot axis number ({0, 1, 2, 4})
            and values should be tuple specifying (``xmin``, ``xmax``). If ``None``, x-axis limits are auto-scaled.

        Returns
        -------
        fig : plt.Figure
            Figure object for evaluation plot
        axes : array of plt.Axes
            Array of Axes objects, each representing a subplot within the figure.

        Notes
        -----
        * The data is ranked in ascending order of the average ranking of the scale sets.

        See Also
        --------
        * :meth:`AAclust.eval` for details on evaluation measures.

        Examples
        --------
        .. include:: examples/aac_plot_eval.rst
        """
        # Check input
        ut.check_df(name="df_eval", df=df_eval, cols_requiered=ut.COLS_EVAL_AACLUST, accept_none=False, accept_nan=False)
        ut.check_dict_xlims(dict_xlims=dict_xlims, n_ax=4)
        ut.check_figsize(figsize=figsize, accept_none=True)
        # Plotting
        colors = ut.plot_get_clist_(n_colors=4)
        fig, axes = plot_eval(df_eval=df_eval,
                              dict_xlims=dict_xlims,
                              figsize=figsize,
                              colors=colors)
        return fig, axes

    def centers(self,
                X: ut.ArrayLike2D,
                labels: ut.ArrayLike1D = None,
                component_x: int = 1,
                component_y: int = 2,
                ax: Optional[plt.Axes] = None,
                figsize: Tuple[Union[int, float], Union[int, float]] = (7, 6),
                legend: bool = True,
                dot_size: int = 100,
                dot_alpha: Union[int, float] = 0.75,
                palette: Optional[mpl.colors.ListedColormap] = None,
                ) -> Tuple[plt.Axes, pd.DataFrame]:
        """
        PCA plot of clustering with centers highlighted

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature matrix. `Rows` typically correspond to scales and `columns` to amino acids.
        labels : array-like, shape (n_samples,)
            Cluster labels for each sample in ``X``. If ``None``, no grouping is used.
        component_x : int, default=1
            Index of the PCA component for the x-axis. Must be >= 1.
        component_y : int, default=2
            Index of the PCA component for the y-axis. Must be >= 1.
        ax : plt.Axes, optional
            Pre-defined Axes object to plot on. If ``None``, a new Axes object is created.
        figsize : tuple, default=(7, 6)
            Figure dimensions (width, height) in inches.
        legend : bool, default=True
            Whether to show the legend.
        dot_size : int, default=100
            Size of the plotted dots.
        dot_alpha : float or int, default=0.75
            The transparency alpha value [0-1] of the plotted dots.
        palette : list, optional
            Colormap for the labels or list of colors. If ``None``, a default colormap is used.

        Returns
        -------
        ax : plt.Axes
            PCA plot axes object.
        df_components : pd.DataFrame
            DataFrame with the PCA components.

        Notes
        -----
        * Ensure `X` and `labels` are in the same order to avoid mislabeling.

        See Also
        --------
        * See the :ref:`tutorial <palette_tutorial>` for more information.
        * See colormaps from matplotlib in :class:`matplotlib.colors.ListedColormap`.

        Examples
        --------
        .. include:: examples/aac_plot_centers.rst
        """
        # Check input
        X = ut.check_X(X=X)
        ut.check_X_unique_samples(X=X)
        labels = ut.check_labels(labels=labels)
        ut.check_match_X_labels(X=X, labels=labels)
        ut.check_number_range(name="component_x", val=component_x, accept_none=False, min_val=1, just_int=True)
        ut.check_number_range(name="component_y", val=component_y, accept_none=False, min_val=1, just_int=True)
        ut.check_ax(ax=ax, accept_none=True)
        ut.check_figsize(figsize=figsize, accept_none=True)
        ut.check_number_range(name="dot_alpha", val=dot_alpha, accept_none=False, min_val=0, max_val=1, just_int=False)
        ut.check_number_range(name="dot_size", val=dot_size, accept_none=False, min_val=1, just_int=True)
        ut.check_bool(name="legend", val=legend)
        ut.check_palette(name="palette", val=palette, accept_none=True)
        # Plotting
        ax, df_components = plot_center_or_medoid(X, labels=labels, plot_centers=True,
                                                  component_x=component_x, component_y=component_y,
                                                  model_class=self._model_class, model_kwargs=self._model_kwargs,
                                                  ax=ax, figsize=figsize,
                                                  dot_size=dot_size, dot_alpha=dot_alpha,
                                                  legend=legend, palette=palette)
        return ax, df_components


    def medoids(self,
                X: ut.ArrayLike2D,
                labels: ut.ArrayLike1D = None,
                component_x: int = 1,
                component_y: int = 2,
                metric: str = "euclidean",
                ax: Optional[plt.Axes] = None,
                figsize: Tuple[Union[int, float], Union[int, float]] = (7, 6),
                legend: bool = True,
                dot_size: int = 100,
                dot_alpha: Union[int, float] = 0.75,
                palette: Optional[mpl.colors.ListedColormap] = None,
                ) -> Tuple[plt.Axes, pd.DataFrame]:
        """
        PCA plot of clustering with medoids highlighted

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature matrix. `Rows` typically correspond to scales and `columns` to amino acids.
        labels : array-like, shape (n_samples,)
            Cluster labels for each sample in ``X``. If ``None``, no grouping is used.
        component_x : int, default=1
            Index of the PCA component for the x-axis. Must be >= 1.
        component_y : int, default=1
            Index of the PCA component for the y-axis. Must be >= 1.
        metric : {'correlation', 'euclidean', 'manhattan', 'cosine'}, default='euclidean'
            The distance metric for calculating medoid.

            - ``correlation``: Pearson correlation (maximum)
            - ``euclidean``: Euclidean distance (minimum)
            - ``manhattan``: Manhattan distance (minimum)
            - ``cosine``: Cosine distance (minimum)

        ax : plt.Axes, optional
            Pre-defined Axes object to plot on. If ``None``, a new Axes object is created.
        figsize : tuple, default=(7, 6)
            Figure dimensions (width, height) in inches.
        legend : bool, default=True
            Whether to show the legend.
        dot_size : int, default=100
            Size of the plotted dots.
        dot_alpha : float or int, default=0.75
            The transparency alpha value [0-1] of the plotted dots.
        palette : list, optional
            Colormap for the labels or list of colors. If ``None``, a default colormap is used.

        Returns
        -------
        ax : plt.Axes
            PCA plot axes object.
        df_components : pd.DataFrame
            DataFrame with the PCA components.

        Notes
        -----
        * Ensure `X` and `labels` are in the same order to avoid mislabeling.

        See Also
        --------
        * See the :ref:`tutorial <palette_tutorial>` for more information.
        * See colormaps from matplotlib in :class:`matplotlib.colors.ListedColormap`.

        Examples
        --------
        .. include:: examples/aac_plot_medoids.rst
        """
        # Check input
        X = ut.check_X(X=X)
        ut.check_X_unique_samples(X=X)
        labels = ut.check_labels(labels=labels)
        ut.check_match_X_labels(X=X, labels=labels)
        ut.check_number_range(name="component_x", val=component_x, accept_none=False, min_val=1, just_int=True)
        ut.check_number_range(name="component_y", val=component_y, accept_none=False, min_val=1, just_int=True)
        check_metric(metric=metric)
        ut.check_ax(ax=ax, accept_none=True)
        ut.check_figsize(figsize=figsize, accept_none=True)
        ut.check_number_range(name="dot_alpha", val=dot_alpha, accept_none=False, min_val=0, max_val=1, just_int=False)
        ut.check_number_range(name="dot_size", val=dot_size, accept_none=False, min_val=1, just_int=True)
        ut.check_bool(name="legend", val=legend)
        ut.check_palette(name="palette", val=palette, accept_none=True)
        # Create plot
        ax, df_components = plot_center_or_medoid(X, labels=labels, plot_centers=False, metric=metric,
                                                  component_x=component_x, component_y=component_y,
                                                  model_class=self._model_class, model_kwargs=self._model_kwargs,
                                                  ax=ax, figsize=figsize,
                                                  dot_size=dot_size, dot_alpha=dot_alpha,
                                                  legend=legend, palette=palette)
        return ax, df_components

    def correlation(self,
                    df_corr: pd.DataFrame = None,
                    labels: ut.ArrayLike1D = None,
                    labels_ref: Optional[ut.ArrayLike1D] = None,
                    cluster_x: bool = False,
                    method: str = "average",
                    xtick_label_rotation: int = 90,
                    ytick_label_rotation: int = 0,
                    bar_position: Union[str, List[str]] = "left",
                    bar_colors: Union[str, List[str]] = "tab:gray",
                    bar_width_x: float = 0.1,
                    bar_spacing_x: float = 0.1,
                    bar_width_y: float = 0.1,
                    bar_spacing_y: float = 0.1,
                    vmin: float = -1.0,
                    vmax: float = 1.0,
                    cmap: str = "viridis",
                    kwargs_heatmap : Optional[dict] = None
                    ) -> plt.Axes:
        """
        Heatmap for correlation matrix with colored sidebar to label clusters.

        Parameters
        ----------
        df_corr : pd.DataFrame, shape (n_samples, n_clusters)
            DataFrame with correlation matrix. `Rows` typically correspond to scales and `columns` to clusters.
        labels : array-like, shape (n_samples,)
            Cluster labels determining the grouping and coloring of the side colorbar.
            It should have the same length as number of rows in ``df_corr`` (n_samples).
        labels_ref  : array-like, shape (n_clusters,), optional
            Cluster labels comprising unique values from 'labels'. Length must match with 'n_clusters' in ``df_corr``.
        cluster_x : bool, default=False
            If ``True``, x-axis (`clusters`) values are clustered. Disabled for pairwise correlation.
        method : {'single', 'complete', 'average', 'weighted', 'centroid', 'median', 'ward'}, default='average'
            Linkage method from :func:`scipy.cluster.hierarchy.linkage` used for clustering.
        xtick_label_rotation : int, default=90
            Rotation of x-tick labels (names of `clusters`).
        ytick_label_rotation : int, default=0
            Rotation of y-tick labels (names of `samples`).
        bar_position : str or list of str, default='left'
            Position of the colored sidebar (``left``, ``right``, ``top``, or ``down``). If ``None``, no sidebar is added.
        bar_colors : str or list of str, default='tab:gray'
            Either a single color or a list of colors for each unique label in ``labels``.
        bar_width_x : float, default=0.1
            Width of the x-axis sidebar, must be >= 0.
        bar_spacing_x : float, default=0.1
            Space between the heatmap and the colored x-axis sidebar, must be >= 0.
        bar_width_y : float, default=0,1
            Width of the y-axis sidebar, must be >= 0.
        bar_spacing_y : float, default=0.1
            Space between the heatmap and the colored y-axis sidebar, must be >= 0.
        vmin : float, default=-1.0
            Minimum value of the color scale in :func:`seaborn.heatmap`.
        vmax : float, default=1.0
            Maximum value of the color scale in :func:`seaborn.heatmap`.
        cmap : str, default='viridis'
            Colormap to be used for the :func:`seaborn.heatmap`.
        kwargs_heatmap : dict, optional
            Dictionary with keyword arguments for adjusting heatmap (:func:`seaborn.heatmap`).

        Returns
        -------
        ax : plt.Axes
            Axes object with the correlation heatmap.

        Notes
        -----
        * Ensure ``labels`` and ``df_corr`` are in the same order to avoid mislabeling.
        * ``bar_tick_labels=True`` will remove tick labels and set them as text for optimal spacing
          so that they can not be adjusted or retrieved afterward (e.g., via `ax.get_xticklabels()`).

        See Also
        --------
        * :func:`seaborn.heatmap`: Seaborn function for creating heatmaps.

        Examples
        --------
        .. include:: examples/aac_plot_correlation.rst
        """
        # Check input
        ut.check_df(name="df_corr", df=df_corr, accept_none=False, accept_nan=False)
        labels = check_match_df_corr_labels(df_corr=df_corr, labels=labels)
        pairwise = [str(x) for x in list(df_corr)] == [str(x) for x in list(df_corr.T)]
        labels_ref = check_match_df_corr_labels_ref(df_corr=df_corr, labels_ref=labels_ref, labels=labels,
                                                    pairwise=pairwise, verbose=self._verbose)
        ut.check_bool(name="cluster_x", val=cluster_x)
        check_match_df_corr_clust_x(df_corr=df_corr, cluster_x=cluster_x)
        check_method(method=method)
        ut.check_number_val(name="xtick_label_rotation", val=xtick_label_rotation, just_int=True, accept_none=True)
        ut.check_number_val(name="ytick_label_rotation", val=ytick_label_rotation, just_int=True, accept_none=True)
        bar_position = check_bar_position(bar_position=bar_position)
        bar_colors = check_bar_colors(bar_colors=bar_colors)
        bar_colors = check_match_bar_colors_labels(bar_colors=bar_colors, labels=labels)
        dict_float_args = dict(bar_width_x=bar_width_x, bar_spacing_x=bar_spacing_x,
                               bar_width_y=bar_width_y, bar_spacing_y=bar_spacing_y)
        for name in dict_float_args:
            ut.check_number_range(name=name, val=dict_float_args[name], min_val=0, accept_none=False, just_int=False)
        ut.check_vmin_vmax(vmin=vmin, vmax=vmax)
        ut.check_cmap(name="cmap", val=cmap, accept_none=False)
        ut.check_dict(name="kwargs_heatmap", val=kwargs_heatmap, accept_none=True)
        # Plotting
        if pairwise:
            cluster_x = False
        try:
            ax = plot_correlation(df_corr=df_corr.copy(), labels=labels, labels_ref=labels_ref,
                                  cluster_x=cluster_x, method=method,
                                  xtick_label_rotation=xtick_label_rotation,
                                  ytick_label_rotation=ytick_label_rotation,
                                  bar_position=bar_position, bar_colors=bar_colors,
                                  bar_width_x=bar_width_x, bar_spacing_x=bar_spacing_x,
                                  bar_width_y=bar_width_y, bar_spacing_y=bar_spacing_y,
                                  vmin=vmin, vmax=vmax, cmap=cmap,
                                  kwargs_heatmap=kwargs_heatmap)
        except Exception as e:
            str_error = f"Following error occurred due to sns.heatmap() function: {e}"
            raise ValueError(str_error)
        plt.tight_layout()
        return ax


"""
This is a script for special statistical measures not available or adjusted from popular python data analysis packages
such as scikit-learn, scipy, or statsmodels.

Developer note: Measures are implemented in aanalysis.utils.metrics to access them within the aanalysis package.
"""
import numpy as np

from aaanalysis.utils import auc_adjusted_, kullback_leibler_divergence_, bic_score_
import aaanalysis.utils as ut

# Helper functions
def _check_n_classes_n_samples(X=None, labels=None):
    """Check matching X and labels"""
    n_classes = len(set(labels))
    n_samples, n_features = X.shape
    if n_classes >= n_samples:
        raise ValueError(f"Number of classes in 'labels' ({n_classes}) must be smaller than n_samples ({n_samples})")
    if n_features == 0:
        raise ValueError(f"'n_features' should not be 0")


# Adjusted Area Under the Curve (AUC*)
def comp_auc_adjusted(X: ut.ArrayLike2D = None,
                      labels: ut.ArrayLike1D = None,
                      label_test: int = 1,
                      label_ref: int = 0,
                      ) -> ut.ArrayLike1D:
    """
    Compute an adjusted Area Under the Curve (AUC) [-0.5, 0.5] assessing the similarity between two groups.

    Introduced in [Breimann24c]_, this adjusted AUC (denoted 'AUC*') is computed for each feature in the
    dataset ``X``, comparing two groups specified by the labels. It is based on the non-parametric measure of the
    difference between two groups. The adjustment of AUC subtracts 0.5, so it ranges between -0.5 and 0.5.
    An AUC* of 0 indicates an equal distribution between the two groups. This measure is useful for ranking features
    based on their ability to distinguish between the two groups.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        Feature matrix. 'Rows' typically correspond to proteins and 'columns' to features.
    labels : array-like, shape (n_samples,)
        Dataset labels of samples in X. Should contain only two different integer label values,
        representing test and reference group (typically, 1 and 0).
    label_test : int, default=1,
        Class label of test group in ``labels``.
    label_ref : int, default=0,
        Class label of reference group in ``labels``.

    Returns
    -------
    auc : array-like, shape (n_features,)
        Array with AUC* values for each feature, ranging from [-0.5, 0.5].
        A value of 0 indicates equal distributions between the two groups for that feature.

    Examples
    --------
    .. include:: examples/comp_auc_adjusted.rst
    """
    # Check input
    X = ut.check_X(X=X, min_n_features=1)
    ut.check_X_unique_samples(X=X, min_n_unique_samples=1)
    ut.check_number_val(name="label_test", val=label_test, just_int=True, accept_none=False)
    ut.check_number_val(name="label_ref", val=label_ref, just_int=True, accept_none=False)
    ut.check_labels(labels=labels, vals_requiered=[label_test, label_ref], allow_other_vals=False)
    ut.check_match_X_labels(X=X, labels=labels)
    # Compute adjusted AUC
    auc = auc_adjusted_(X=X, labels=labels, label_test=label_test)
    return auc


# BIC score
def comp_bic_score(X: ut.ArrayLike2D = None,
                   labels: ut.ArrayLike1D =None
                   ) -> float:
    """
    Compute an adjusted Bayesian Information Criterion (BIC) (-∞, ∞) for assessing clustering quality.

    Described in [Breimann24b], this adjusted BIC is computed for a given set of clusters in the dataset ``X``.
    The BIC is a clustering model selection criterion that balances the model complexity against the
    likelihood of the data distribution. Unlike the traditional BIC where lower values are better, this adjusted BIC,
    is modified to align with other clustering evaluation measures like the
    Silhouette coefficient and the Calinski-Harabasz score. In this adjusted version, higher values indicate
    better clustering.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        Feature matrix. 'Rows' typically correspond to proteins and 'columns' to features.
    labels : array-like, shape (n_samples,)
        Predicted labels for each sample. Each label corresponds to a cluster.

    Returns
    -------
    bic : float
        The Bayesian Information Criterion value. A lower BIC value indicates a better model fit to the data.

    Notes
    -----
    *  An `epsilon` value (1e-10) is utilized to prevent division by zero in the computation.

    See Also
    --------
    * The Silhouette coefficient [-1, 1] can be computed by :func:`sklearn.metrics.silhouette_score`.
    * The Calinski Harabasz score [0, ∞] can be obtained using :func:`sklearn.metrics.calinski_harabasz_score`.
    * Clustering evaluation can be performed using :meth:`AAclust.eval`.

    Examples
    --------
    .. include:: examples/comp_bic_score.rst
    """
    # Check input
    X = ut.check_X(X=X, min_n_features=1)
    ut.check_X_unique_samples(X=X)
    ut.check_labels(labels=labels)
    ut.check_match_X_labels(X=X, labels=labels)
    _check_n_classes_n_samples(X=X, labels=labels)
    # Compute bic
    bic = bic_score_(X, labels=labels)
    return bic


# Kullback-Leibler Divergence
def comp_kld(X: ut.ArrayLike2D = None,
             labels: ut.ArrayLike1D =None,
             label_test: int = 1,
             label_ref: int = 0
             ) -> ut.ArrayLike1D:
    """
    Compute the Kullback-Leibler Divergence (KLD) [0, ∞) for assessing the similarity between two groups.

    The KLD is calculated for each feature in ``X``, comparing the distributions between two subgroups specified
    by ``label_test`` and ``label_ref`` in labels. Generally, the KLD measures how one probability distribution
    diverges from a second, expected probability distribution. Higher KLD values indicate more divergence. The observed
    upper limit lies around 200 indicating complete divergence of two non-overlapping distributions.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        Feature matrix. `Rows` typically correspond to proteins and `columns` to features.
    labels : array-like, shape (n_samples,)
        Labels for each sample in ``X``. Should contain only integer label values and at least 2 per class.
    label_test : int, default=1,
        Class label of test group in ``labels``.
    label_ref : int, default=0,
        Class label of reference group in ``labels``.

    Returns
    -------
    kld : array-like, shape (n_features,)
        Array of Kullback-Leibler Divergence values for each feature in ``X``. Each value represents
        the divergence of the test group distribution from the reference group distribution for that feature.

    Notes
    -----
    * For valid KLD calculations, the input matrix `X` must meet certain conditions:

      - Ensure adequate variability of features in ``X`` to avoid computational problems like singular
        covariance matrices in Gaussian KDE.
      - Avoid rows in ``X`` lying in a lower-dimensional subspace; consider dimensionality reduction if necessary.

    See Also
    --------
    * :func:`scipy.stats.gaussian_kde` function representing a kernel-density estimate using Gaussian kernels.
      It is used for estimating the probability density function of a random variable (i.e., feature in ``X``),
      which is a crucial step in the computation of Kullback-Leibler Divergence (KLD).
    * :func:`scipy.stats.entropy` function for computing the Shannon entropy. In the context of KLD,
      it is used to measure the divergence between two probability distributions, typically derived
      from kernel-density estimates of different data groups.

    Examples
    --------
    .. include:: examples/comp_kld.rst
    """
    # Check input
    X = ut.check_X(X=X, min_n_features=1)
    ut.check_X_unique_samples(X=X, min_n_unique_samples=3)
    ut.check_number_val(name="label_test", val=label_test, just_int=True, accept_none=False)
    ut.check_number_val(name="label_ref", val=label_ref, just_int=True, accept_none=False)
    ut.check_labels(labels=labels, vals_requiered=[label_test, label_ref],
                    n_per_group_requiered=2, allow_other_vals=False)
    ut.check_match_X_labels(X=X, labels=labels, check_variability=True)
    # Compute tge Kullback-Leibler divergence
    try:
        kld = kullback_leibler_divergence_(X=X, labels=labels, label_test=label_test, label_ref=label_ref)
    except Exception as e:
        raise ValueError(f"Following error occurred during the computation of Kullback-Leibler Divergence: {e}")
    return kld



"""
This is a script for the backend of the CPPPlot.update_seq_size() method.
"""
import matplotlib.pyplot as plt
from ._utils_cpp_plot_positions import PlotPartPositions

# I Helper functions
def _get_sorted_x_tick_labels(ax=None):
    # Get all x-axis tick labels
    labels = ax.xaxis.get_ticklabels(which="both")
    # Compute the positions of the labels and sort them
    f = lambda l: l.get_window_extent(ax.figure.canvas.get_renderer())
    tick_positions = [f(l).x0 for l in labels]
    sorted_tick_positions, sorted_labels = zip(*sorted(zip(tick_positions, labels), key=lambda t: t[0]))
    return sorted_labels


# II Main Function
def get_tmd_jmd_seq(ax=None, jmd_n_len=10, jmd_c_len=10):
    """Get tmd_seq, jmd_n_seq, and jmd_c_seq from axes"""
    _labels = _get_sorted_x_tick_labels(ax=ax)
    tmd_jmd_seq = "".join([x.get_text() for x in _labels])
    jmd_n_seq = tmd_jmd_seq[0:jmd_n_len]
    tmd_seq = tmd_jmd_seq[jmd_n_len:-jmd_c_len]
    jmd_c_seq = tmd_jmd_seq[-jmd_c_len:]
    return jmd_n_seq, tmd_seq, jmd_c_seq


def update_seq_size_(ax=None, tmd_seq=None, jmd_n_seq=None, jmd_c_seq=None,
                     max_x_dist=0.1,
                     tmd_color="mediumspringgreen", jmd_color="blue",
                     tmd_seq_color="black", jmd_seq_color="white"):
    """Update the font size of the sequence characters to prevent overlap."""
    colors = [jmd_color] * len(jmd_n_seq) + [tmd_color] * len(tmd_seq) + [jmd_color] * len(jmd_c_seq)
    dict_seq_color = {tmd_color: tmd_seq_color, jmd_color: jmd_seq_color}
    # Get all x-axis tick labels
    labels = _get_sorted_x_tick_labels(ax=ax)
    # Adjust font size to prevent overlap
    pp = PlotPartPositions()
    seq_size = pp.get_optimal_fontsize(ax=ax, labels=labels, max_x_dist=max_x_dist)
    lw = plt.gcf().get_size_inches()[0]/5
    for l, c in zip(labels, colors):
        l.set_fontsize(seq_size)
        l.set_bbox(dict(facecolor=c, edgecolor=c, zorder=0.1, alpha=1,
                        clip_on=False,
                        pad=0, linewidth=lw))
        l.set_color(dict_seq_color[c])
    return ax, seq_size


def update_tmd_jmd_labels(fig=None, seq_size=None, fontsize_tmd_jmd=None, weight_tmd_jmd="bold"):
    """Adjust size and position of TMD-JMD labels"""
    fs_labels = fontsize_tmd_jmd if fontsize_tmd_jmd else seq_size
    if fig is not None and len(fig.axes) > 1:
        ax2 = fig.axes[1]
        ax2.spines["bottom"].set_position(("outward", seq_size + 7))
        ax2.tick_params(axis='x', labelsize=fs_labels)
        for label in ax2.get_xticklabels():
            label.set_weight(weight_tmd_jmd)
